# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Do the conversion of search spaces into a normalized cartesian space.
"""
import numpy as np
from scipy.interpolate import interp1d
from scipy.special import expit as logistic  # because nobody calls it expit
from scipy.special import logit

from bayesmark.np_util import clip_chk, snap_to

WARPED_DTYPE = np.float_
N_GRID_DEFAULT = 8

# I can't make up mind of unicode or str is better wrt to Py 2/3 compatibility
# ==> Just make a global constant and make sure it works either way.
# Note: if we switch to np.str_, we will also need to update doc-strings!
CAT_DTYPE = np.unicode_
CAT_KIND = "U"
CAT_NATIVE_DTYPE = str
# Check to make sure consistent
assert CAT_KIND == np.dtype(CAT_DTYPE).kind
_infered = type(CAT_DTYPE("").item())
assert CAT_NATIVE_DTYPE == _infered

# ============================================================================
# These could go into util
# ============================================================================


def unravel_index(dims):
    """Builds tuple of coordinate arrays to traverse an `numpy` array.

    Wrapper around :func:`numpy:numpy.unravel_index` that avoids bug at corner case for ``dims=()``. The fix for this
    has been merged into the numpy master branch Oct 18, 2017 so future numpy releases will make this wrapper not
    needed. Otherwise, ``unravel_index(X.shape)`` is equivalent to: ``np.unravel_index(range(X.size), X.shape)``.

    Parameters
    ----------
    dims : tuple(int)
        The shape of the array to use for unraveling ``indices``.

    Returns
    -------
    unraveled_coords : tuple(:class:`numpy:numpy.ndarray`)
        Each array in the tuple has shape (n,) where ``n=np.prod(dims)``.

    References
    ----------
    unravel_index(0, ()) should return () (Trac #2120) #580
    https://github.com/numpy/numpy/issues/580
    Allow `unravel_index(0, ())` to return () #9884
    https://github.com/numpy/numpy/pull/9884
    """
    size = np.prod(dims)
    if dims == () or size == 0:  # The corner case
        return ()

    idx = np.unravel_index(range(np.prod(dims)), dims)
    return idx


def encode(X, labels, assume_sorted=False, dtype=bool, assume_valid=False):
    """Perform one hot encoding of categorical data in :class:`numpy:numpy.ndarray` variable `X` of any dimension.

    Parameters
    ----------
    X : :class:`numpy:numpy.ndarray` of shape (...)
        Categorical values of any standard type. Vectorized to work for any dimensional `X`.
    labels : :class:`numpy:numpy.ndarray` of shape (n,)
        Complete list of all possible labels. List is flattened if it is not already 1 dimensional.
    assume_sorted : bool
        If true, assume labels is already sorted and unique. This saves the computational cost of calling
        :func:`numpy:numpy.unique`.
    dtype : type
        Desired data of feature array. One-hot is most logically `bool`, but feature matrices are usually `float`.
    assume_valid : bool
        If true, assume all element of `X` are in the list `labels`. This saves the computational cost of verifying
        `X` are in `labels`. If true and a non-label `X` occurs this routine will silently give bogus result.

    Returns
    -------
    Y : :class:`numpy:numpy.ndarray` of shape (..., n)
        One-hot encoding of `X`. Extra dimension is appended at end for the one-hot vector. It has data type `dtype`.
    """
    X = np.asarray(X)
    labels = np.asarray(labels) if assume_sorted else np.unique(labels)
    check_array(labels, "labels", pre=True, ndim=1, min_size=1)

    idx = np.searchsorted(labels, X)
    # If x is not even in labels then this will fail. This is not ValueError
    # because the user explictly asked for this using argument assume_valid.
    assert assume_valid or np.all(np.asarray(labels[idx]) == X)

    # This is using some pro np indexing technique to vectorize across all
    # possible input dimensions for X in the same code.
    Y = np.zeros(X.shape + (len(labels),), dtype=dtype)
    Y[unravel_index(X.shape) + (idx.ravel(),)] = True
    return Y


def decode(Y, labels, assume_sorted=False):
    """Perform inverse of one-hot encoder `encode`.

    Parameters
    ----------
    Y : :class:`numpy:numpy.ndarray` of shape (..., n)
        One-hot encoding of categorical data `X`. Extra dimension is appended at end for the one-hot vector. Maximum
        element is taken if there is more than one non-zero entry in one-hot vector.
    labels : :class:`numpy:numpy.ndarray` of shape (n,)
        Complete list of all possible labels. List is flattened if it is not already 1-dimensional.
    assume_sorted : bool
        If true, assume labels is already sorted and unique. This saves the computational cost of calling
        :func:`numpy:numpy.unique`.

    Returns
    -------
    X : :class:`numpy:numpy.ndarray` of shape (...)
        Categorical values corresponding to one-hot encoded `Y`.
    """
    Y = np.asarray(Y)
    labels = np.asarray(labels) if assume_sorted else np.unique(labels)
    check_array(labels, "labels", pre=True, ndim=1, min_size=1)
    check_array(Y, "Y", pre=True, shape_endswith=(len(labels),))

    idx = np.argmax(Y, axis=-1)
    X = labels[idx]
    return X


def _error(msg, pre=False):  # pragma: validator
    """Helper routine for :func:`.check_array`.

    This could probably be made cleaner by using raise to create the assert.
    """
    if pre:
        raise ValueError(msg)
    else:
        assert False, msg


def check_array(
    X,
    name,
    pre=False,
    ndim=None,
    shape=None,
    shape_endswith=(),
    min_size=0,
    dtype=None,
    kind=None,
    allow_infinity=True,
    allow_nan=True,
    unsorted=True,
    whitelist=None,
):  # pragma: validator
    """Like :func:`sklearn:sklearn.utils.check_array` but better.

    Check specified property of input array `X`. If an argument is not specified it passes by default.

    Parameters
    ----------
    X : :class:`numpy:numpy.ndarray`
        `numpy` array we want to validate.
    name : str
        Human readable name of of variable to refer to it in error messages. Note this can include spaces unlike simply
        using the variable name.
    pre : bool
        If true, interpret this as check as validating pre-conditions to a function and will raise a `ValueError` if a
        check fails. If false, assumes we are checking post-conditions and will raise an assertion failure.
    ndim : int
        Expected value of ``X.ndim``.
    shape : tuple(int)
        Expected value of ``X.shape``.
    shape_endswith : tuple(int)
        Expected that ``X.shape`` ends with `shape_endswith`. This is useful in broadcasting where extra dimensions
        might be added on.
    min_size : int
        Minimum value for ``X.size``
    dtype : dtype
        Expected value of ``X.dtype``.
    kind : str
        Expected value of ``X.dtype.kind``. This is `'f'` for `float`, `'i'` for `int`, and so on.
    allow_infinity : bool
        If false, the check fails when `X` contains inf or ``-inf``.
    allow_nan : bool
        If false, the check fails when `X` contains a ``NaN``.
    unsorted : bool
        If false, the check fails when `X` is not in sorted order. This is designed to even work with string arrays.
    whitelist : :class:`numpy:numpy.ndarray`
        Array containing allowed values for `X`. If an element of `X` is not found in `whitelist`, the check fails.
    """
    if (ndim is not None) and X.ndim != ndim:
        _error("Expected %d dimensions for %s, got %d" % (ndim, name, X.ndim), pre)

    if (shape is not None) and X.shape != shape:
        _error("Expected shape %s for %s, got %s" % (str(shape), name, str(X.shape)), pre)

    if len(shape_endswith) > 0:
        if X.shape[-len(shape_endswith) :] != shape_endswith:
            if len(shape_endswith) == 1:
                _error("Expected shape (..., %d) for %s, got %s" % (shape_endswith[0], name, str(X.shape)), pre)
            else:
                _error("Expected shape (..., %s for %s, got %s" % (str(shape_endswith)[1:], name, str(X.shape)), pre)

    if (min_size > 0) and (X.size < min_size):
        _error("%s needs at least %d elements, it has %d" % (name, min_size, X.size), pre)

    if (dtype is not None) and X.dtype != np.dtype(dtype):
        _error("Expected dtype %s for %s, got %s" % (str(np.dtype(dtype)), name, str(X.dtype)), pre)

    if (kind is not None) and X.dtype.kind != kind:
        _error("Expected array with kind %s for %s, got %s" % (kind, name, str(X.dtype.kind)), pre)

    if (not allow_infinity) and np.any(np.abs(X) == np.inf):
        _error("Infinity is not allowed in %s" % name, pre)

    if (not allow_nan) and np.any(np.isnan(X)):
        _error("NaN is not allowed in %s" % name, pre)

    if whitelist is not None:
        ok = np.all([xx in whitelist for xx in np.nditer(X, ["zerosize_ok"])])
        if not ok:
            _error("Expected all elements of %s to be in %s" % (name, str(whitelist)), pre)

    # Only do this check in 1D
    if X.ndim == 1 and (not unsorted) and np.any(X[:-1] > X[1:]):
        _error("Expected sorted input for %s" % name, pre)


# ============================================================================
# Setup warping dictionaries
# ============================================================================


def identity(x):
    """Helper function that perform warping in linear space. Sort of a no-op.

    Parameters
    ----------
    x : scalar
        Input variable in linear space. Can be any numeric type and is vectorizable.

    Returns
    -------
    y : scalar
        Same as input `x`.
    """
    y = x
    return y


def bilog(x):
    """Bilog warping function. Extension of log to work with negative numbers.

    ``Bilog(x) ~= log(x)`` for large `x` or ``-log(abs(x))`` if `x` is negative. However, the bias term ensures good
    behavior near 0 and ``bilog(0) = 0``.

    Parameters
    ----------
    x : scalar
        Input variable in linear space. Can be any numeric type and is vectorizable.

    Returns
    -------
    y : float
        The bilog of `x`.
    """
    y = np.sign(x) * np.log(1.0 + np.abs(x))
    return y


def biexp(x):
    """Inverse of :func:`.bilog` function.

    Parameters
    ----------
    x : scalar
        Input variable in linear space. Can be any numeric type and is vectorizable.

    Returns
    -------
    y : float
        The biexp of `x`.
    """
    y = np.sign(x) * (np.exp(np.abs(x)) - 1.0)
    return y


WARP_DICT = {"linear": identity, "log": np.log, "logit": logit, "bilog": bilog}
UNWARP_DICT = {"linear": identity, "log": np.exp, "logit": logistic, "bilog": biexp}

# ============================================================================
# Setup spaces class hierarchy
# ============================================================================


class Space(object):
    """Base class for all types of variables.
    """

    def __init__(self, dtype, default_round, warp="linear", values=None, range_=None):
        """Generic constructor of `Space` class.

        Not intended to be called directly but instead by child classes. However, `Space` is not an abstract class and
        will not give an error when instantiated.
        """
        self.dtype = dtype
        assert warp in WARP_DICT, "invalid space %s, allowed spaces are: %s" % (str(warp), str(WARP_DICT.keys()))
        self.warp_f = WARP_DICT[warp]
        self.unwarp_f = UNWARP_DICT[warp]

        # Setup range and rounding if values is suplied
        assert (values is None) != (range_ is None)
        round_to_values = default_round
        if range_ is None:  # => value is not None
            # Debatable if unique should be done before or after cast. But I
            # think after is better, esp. when changing precisions.
            values = np.asarray(values, dtype=dtype)
            values = np.unique(values)  # values now 1D ndarray no matter what
            check_array(
                values,
                "unique values",
                pre=True,
                ndim=1,
                dtype=dtype,
                min_size=2,
                allow_infinity=False,
                allow_nan=False,
            )

            # Extrapolation might happen due to numerics in type conversions.
            # Bounds checking is still done in validate routines.
            round_to_values = interp1d(values, values, kind="nearest", fill_value="extrapolate")
            range_ = (values[0], values[-1])
        # Save values and rounding
        # Values is either None or was validated inside if statement
        self.values = values
        self.round_to_values = round_to_values

        # Note that if dtype=None that is the default for asarray.
        range_ = np.asarray(range_, dtype=dtype)
        check_array(range_, "range", pre=True, shape=(2,), dtype=dtype, unsorted=False)
        # Save range info, with input validation and post validation
        self.lower, self.upper = range_

        # Convert to warped bounds too with lots of post validation
        self.lower_warped, self.upper_warped = self.warp_f(range_[..., None]).astype(WARPED_DTYPE, copy=False)
        check_array(
            self.lower_warped,
            "warped lower bound %s(%.1f)" % (warp, self.lower),
            ndim=1,
            pre=True,
            dtype=WARPED_DTYPE,
            allow_infinity=False,
            allow_nan=False,
        )
        # Should never happen if warpers are strictly monotonic:
        assert np.all(self.lower_warped <= self.upper_warped)

        # Make sure a bit bigger to keep away from lower due to numerics
        self.upper_warped = np.maximum(self.upper_warped, np.nextafter(self.lower_warped, np.inf))
        check_array(
            self.upper_warped,
            "warped upper bound %s(%.1f)" % (warp, self.upper),
            pre=True,
            shape=self.lower_warped.shape,
            dtype=WARPED_DTYPE,
            allow_infinity=False,
            allow_nan=False,
        )
        # Should never happen if warpers are strictly monotonic:
        assert np.all(self.lower_warped < self.upper_warped)

    def validate(self, X, pre=False):
        """Routine to validate inputs to warp.

        This routine does not perform any checking on the dimensionality of `X` and is fully vectorized.
        """
        X = np.asarray(X, dtype=self.dtype)

        if self.values is None:
            X = clip_chk(X, self.lower, self.upper)
        else:
            check_array(X, "X", pre=pre, whitelist=self.values)

        return X

    def validate_warped(self, X, pre=False):
        """Routine to validate inputs to unwarp. This routine is vectorized, but `X` must have at least 1-dimension.
        """
        X = np.asarray(X, dtype=WARPED_DTYPE)
        check_array(X, "X", pre=pre, shape_endswith=(len(self.lower_warped),))

        X = clip_chk(X, self.lower_warped, self.upper_warped)
        return X

    def warp(self, X):
        """Warp inputs to a continuous space.

        Parameters
        ----------
        X : :class:`numpy:numpy.ndarray` of shape (...)
            Input variables to warp. This is vectorized to work in any dimension, but it must have the same type code
            as the class, which is in `self.type_code`.

        Returns
        -------
        X_w : :class:`numpy:numpy.ndarray` of shape (..., m)
            Warped version of input space. By convention there is an extra dimension on warped array.
            Currently, ``m=1`` for all warpers. `X_w` will have a `float` type.
        """
        X = self.validate(X, pre=True)

        X_w = self.warp_f(X)
        X_w = X_w[..., None]  # Convention is that warped has extra dim

        X_w = self.validate_warped(X_w)  # Ensures of WAPRED_DTYPE
        check_array(X_w, "X", ndim=X.ndim + 1, dtype=WARPED_DTYPE)
        return X_w

    def unwarp(self, X_w):
        """Inverse of `warp` function.

        Parameters
        ----------
        X_w : :class:`numpy:numpy.ndarray` of shape (..., m)
            Warped version of input space. This is vectorized to work in any dimension. But, by convention, there is an
            extra dimension on the warped array. Currently, the last dimension ``m=1`` for all warpers. `X_w` must be of
            a `float` type.

        Returns
        -------
        X : :class:`numpy:numpy.ndarray` of shape (...)
            Unwarped version of `X_w`. `X` will have the same type code as the class, which is in `self.type_code`.
        """
        X_w = self.validate_warped(X_w, pre=True)

        X = clip_chk(self.unwarp_f(X_w[..., 0]), self.lower, self.upper)
        X = self.round_to_values(X)

        X = self.validate(X)  # Ensures of dtype
        check_array(X, "X", ndim=X_w.ndim - 1, dtype=self.dtype)
        return X

    def get_bounds(self):
        """Get bounds of the warped space.

        Returns
        -------
        bounds : :class:`numpy:numpy.ndarray` of shape (D, 2)
            Bounds in the warped space. First column is the lower bound and the second column is the upper bound.
            Calling ``bounds.tolist()`` gives the bounds in the standard form expected by `scipy` optimizers:
            ``[(lower_1, upper_1), ..., (lower_n, upper_n)]``.
        """
        bounds = np.stack((self.lower_warped, self.upper_warped), axis=1)
        check_array(bounds, "bounds", shape=(len(self.lower_warped), 2), dtype=WARPED_DTYPE)
        return bounds

    def grid(self, max_interp=N_GRID_DEFAULT):
        """Return grid spanning the original (unwarped) space.

        Parameters
        ----------
        max_interp : int
            The number of points to use in grid space when a range and not values are used to define the space.
            Must be ``>= 0``.

        Returns
        -------
        values : list
            Grid spanning the original space. This is simply `self.values` if a grid has already been specified,
            otherwise it is just grid across the range.
        """
        values = self.values
        if values is None:
            vw = np.linspace(self.lower_warped, self.upper_warped, max_interp)
            # Some spaces like int make result in duplicates after unwarping
            # so we apply unique to avoid this. However this will usually be
            # wasted computation.
            values = np.unique(self.unwarp(vw[:, None]))
            check_array(values, "values", ndim=1, dtype=self.dtype)

        # Best to convert to list to make sure in native type
        values = values.tolist()
        return values


class Real(Space):
    """Space for transforming real variables to normalized space (after warping).
    """

    def __init__(self, warp="linear", values=None, range_=None):
        """Build Real space class.

        Parameters
        ----------
        warp : {'linear', 'log', 'logit', 'bilog'}
            Which warping type to apply to the space. The warping is applied in the original space. That is, in a space
            with ``warp='log'`` and ``range_=(2.0, 10.0)``, the value 2.0 warps to ``log(2)``, not ``-inf`` as in some
            other frameworks.
        values : None or list(float)
            Possible values for space to take. Values must be of `float` type.
        range_ : None or :class:`numpy:numpy.ndarray` of shape (2,)
            Array with (lower, upper) pair with limits of space. Note that one must specify `values` or `range_`, but
            not both. `range_` must be composed of `float`.
        """
        assert warp is not None, "warp/space not specified for real"
        Space.__init__(self, np.float_, identity, warp, values, range_)


class Integer(Space):
    """Space for transforming integer variables to continuous normalized space.
    """

    def __init__(self, warp="linear", values=None, range_=None):
        """Build Integer space class.

        Parameters
        ----------
        warp : {'linear', 'log', 'bilog'}
            Which warping type to apply to the space. The warping is applied in the original space. That is, in a space
            with ``warp='log'`` and ``range_=(2, 10)``, the value 2 warps to ``log(2)``, not ``-inf`` as in some other
            frameworks. There are no settings with integers that are compatible with the logit warp.
        values : None or list(float)
            Possible values for space to take. Values must be of `int` type.
        range_ : None or :class:`numpy:numpy.ndarray` of shape (2,)
            Array with (lower, upper) pair with limits of space. Note that one must specify `values` or `range_`, but
            not both. `range_` must be composed of `int`.
        """
        assert warp is not None, "warp/space not specified for int"
        Space.__init__(self, np.int_, np.round, warp, values, range_)


class Boolean(Space):
    """Space for transforming Boolean variables to continuous normalized space.
    """

    def __init__(self, warp=None, values=None, range_=None):
        """Build Boolean space class.

        Parameters
        ----------
        warp : None
            Must be omitted or None, provided for consitency with other types.
        values : None
            Must be omitted or None, provided for consitency with other types.
        range_ : None
            Must be omitted or None, provided for consitency with other types.
        """
        assert warp is None, "cannot warp bool"
        assert (values is None) and (range_ is None), "cannot pass in values or range for bool"
        self.dtype = np.bool_
        self.warp_f = identity
        self.unwarp_f = identity

        self.values = np.array([False, True], dtype=np.bool_)
        self.round_to_values = np.round

        self.lower, self.upper = self.dtype(False), self.dtype(True)
        self.lower_warped = np.array([0.0], dtype=WARPED_DTYPE)
        self.upper_warped = np.array([1.0], dtype=WARPED_DTYPE)


class Categorical(Space):
    """Space for transforming categorical variables to continuous normalized space.
    """

    def __init__(self, warp=None, values=None, range_=None):
        """Build Integer space class.

        Parameters
        ----------
        warp : None
            Must be omitted or None, provided for consitency with other types.
        values : list(str)
            Possible values for space to take. Values must be unicode strings. Requiring type unicode (``'U'``) rather
            than strings (``'S'``) corresponds to the native string type.
        range_ : None
            Must be omitted or None, provided for consitency with other types.
        """
        assert warp is None, "cannot warp cat"
        assert values is not None, "must pass in explicit values for cat"
        assert range_ is None, "cannot pass in range for cat"

        values = np.unique(values)  # values now 1D ndarray no matter what
        check_array(values, "values", pre=True, ndim=1, kind=CAT_KIND, min_size=2)
        self.values = values

        self.dtype = CAT_DTYPE
        # Debatable if decode should go in unwarp or round_to_values
        self.warp_f = lambda x: encode(x, self.values, True, WARPED_DTYPE, True)
        self.unwarp_f = identity
        self.round_to_values = lambda y: decode(y, self.values, True)

        self.lower, self.upper = None, None  # Don't need them
        self.lower_warped = np.zeros(len(values), dtype=WARPED_DTYPE)
        self.upper_warped = np.ones(len(values), dtype=WARPED_DTYPE)

    def warp(self, X):
        """Warp inputs to a continuous space.

        Parameters
        ----------
        X : :class:`numpy:numpy.ndarray` of shape (...)
            Input variables to warp. This is vectorized to work in any dimension, but it must have the same
            type code as the class, which is unicode (``'U'``) for the :class:`.Categorical` space.

        Returns
        -------
        X_w : :class:`numpy:numpy.ndarray` of shape (..., m)
            Warped version of input space. By convention there is an extra dimension on warped array. The warped space
            has a one-hot encoding and therefore `m` is the number of possible values in the space. `X_w` will have
            a `float` type.
        """
        X = self.validate(X, pre=True)

        X_w = self.warp_f(X)

        # Probably over kill to validate here too, but why not:
        X_w = self.validate_warped(X_w)
        check_array(X_w, "X", ndim=X.ndim + 1, dtype=WARPED_DTYPE)
        return X_w

    def unwarp(self, X_w):
        """Inverse of `warp` function.

        Parameters
        ----------
        X_w : :class:`numpy:numpy.ndarray` of shape (..., m)
            Warped version of input space. The warped space has a one-hot encoding and therefore `m` is the number of
            possible values in the space. `X_w` will have a `float` type. Non-zero/one values are allowed in `X_w`.
            The maximal element in the vector is taken as the encoded value.

        Returns
        -------
        X : :class:`numpy:numpy.ndarray` of shape (...)
            Unwarped version of `X_w`. `X` will have same type code as the :class:`.Categorical` class, which is
            unicode (``'U'``).
        """
        X_w = self.validate_warped(X_w, pre=True)

        X = self.round_to_values(self.unwarp_f(X_w))

        X = self.validate(X)
        check_array(X, "X", ndim=X_w.ndim - 1, kind=CAT_KIND)
        return X


# Treat ordinal identically to categorical for now
SPACE_DICT = {"real": Real, "int": Integer, "bool": Boolean, "cat": Categorical, "ordinal": Categorical}

# ============================================================================
# Setup code for joint spaces over multiple parameters with different configs
# ============================================================================


class JointSpace(object):
    """Combination of multiple :class:`.Space` objectives to transform multiple variables at the same time (jointly).
    """

    def __init__(self, meta):
        """Build Real space class.

        Parameters
        ----------
        meta : dict(str, dict)
            Configuration of variables in joint space. See API description.
        """
        assert len(meta) > 0  # Unclear what to do with empty space

        # Lock in an order if not ordered dict, sorted helps reproducibility
        self.param_list = sorted(meta.keys())

        # Might as well pre-validate a bit here
        for param, config in meta.items():
            assert config["type"] in SPACE_DICT, "invalid input type %s" % config["type"]

        spaces = {
            param: SPACE_DICT[config["type"]](
                config.get("space", None), config.get("values", None), config.get("range", None)
            )
            for param, config in meta.items()
        }
        self.spaces = spaces

        self.blocks = np.cumsum([len(spaces[param].get_bounds()) for param in self.param_list])

    def validate(self, X):
        """Raise `ValueError` if X does not match the format expected for a
        joint space."""
        for record in X:
            if self.param_list != sorted(record.keys()):
                raise ValueError("Expected joint space keys %s, but got %s", (self.param_list, sorted(record.keys())))
            for param in self.param_list:
                self.spaces[param].validate([record[param]], pre=True)
        # Return X back so we have option to cast it to list or whatever later
        return X

    def warp(self, X):
        """Warp inputs to a continuous space.

        Parameters
        ----------
        X : list(dict(str, object)) of shape (n,)
            List of `n` points in the joint space to warp. Each list element is a dictionary where each key corresponds
            to a variable in the joint space. Keys can be be missing in the records and the according warped variables
            will be ``nan``.

        Returns
        -------
        X_w : :class:`numpy:numpy.ndarray` of shape (n, m)
            Warped version of input space. Result is 2D `float` np array. `n` is the number of input points, length
            of `X`. `m` is the size of the joint warped space, which can be inferred by calling :func:`.get_bounds`.
        """
        # It would be nice to have cleaner way to deal with this corner case
        if len(X) == 0:
            return np.zeros((0, self.blocks[-1]), dtype=WARPED_DTYPE)

        X_w = [
            np.concatenate(
                [
                    self.spaces[param].warp(record[param])
                    if param in record
                    else np.full(len(self.spaces[param].get_bounds()), np.nan)
                    for param in self.param_list
                ]
            )
            for record in X
        ]
        X_w = np.stack(X_w, axis=0)
        check_array(X_w, "X", shape=(len(X), self.blocks[-1]), dtype=WARPED_DTYPE)
        return X_w

    def unwarp(self, X_w, fixed_vals={}):
        """Inverse of :func:`.warp`.

        Parameters
        ----------
        X_w : :class:`numpy:numpy.ndarray` of shape (n, m)
            Warped version of input space. Must be 2D `float` :class:`numpy:numpy.ndarray`. `n` is the number of
            separate points in the warped joint space. `m` is the size of the joint warped space, which can be inferred
            in advance by calling :func:`.get_bounds`.
        fixed_vals : dict
            Subset of variables we want to keep fixed in X. Unwarp checks that the unwarped version of `X_w` matches
            `fixed_vals` up to numerical error. Otherwise, an error is raised.

        Returns
        -------
        X : list(dict(str, object)) of shape (n,)
            List of `n` points in the joint space to warp. Each list element is a dictionary where each key corresponds
            to a variable in the joint space.
        """
        X_w = np.asarray(X_w)
        check_array(X_w, "X", ndim=2, shape_endswith=(self.blocks[-1],), dtype=WARPED_DTYPE)
        N = X_w.shape[0]

        # Use snap_to to make sure we get exact value (no-round off) for cases where we know expected answer
        X = {
            param: snap_to(self.spaces[param].unwarp(xx), fixed_vals.get(param, None))
            for param, xx in zip(self.param_list, np.hsplit(X_w, self.blocks[:-1]))
        }
        # Convert dict of arrays to list of dicts, this would not be needed if
        # we used pandas but we do not want to add it as a dep. np.asscalar and
        # .item() appear to be the same thing but asscalar seems more readable.
        X = [{param: X[param][ii].item() for param in self.param_list} for ii in range(N)]
        return X

    def get_bounds(self):
        """Get bounds of the warped joint space.

        Returns
        -------
        bounds : :class:`numpy:numpy.ndarray` of shape (m, 2)
            Bounds in the warped space. First column is the lower bound and the second column is the upper bound.
            ``bounds.tolist()`` gives the bounds in the standard form expected by scipy optimizers:
            ``[(lower_1, upper_1), ..., (lower_n, upper_n)]``.
        """
        bounds = np.concatenate([self.spaces[param].get_bounds() for param in self.param_list], axis=0)
        check_array(bounds, "bounds", shape_endswith=(2,), dtype=WARPED_DTYPE)
        return bounds

    def grid(self, max_interp=N_GRID_DEFAULT):
        """Return grid spanning the original (unwarped) space.

        Parameters
        ----------
        max_interp : int
            The number of points to use in grid space when a range and not values are used to define the space.
            Must be ``>= 0``.

        Returns
        -------
        axes : dict(str, list)
            Grids spanning the original spaces of each variable. For each variable, this is simply ``self.values``
            if a grid has already been specified, otherwise it is just grid across the range.
        """
        axes = {var_name: space.grid(max_interp=max_interp) for var_name, space in self.spaces.items()}
        return axes



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Routines to compute and compare the "signatures" of objective functions. These are useful to make sure two different
studies were actually optimizing the same objective function (even if they say the same test case in the meta-data).
"""
import warnings

import numpy as np
import pandas as pd

import bayesmark.random_search as rs

# How many points to probe the function to get the signature
N_SUGGESTIONS = 5


def get_func_signature(f, api_config):
    """Get the function signature for an objective function in an experiment.

    Parameters
    ----------
    f : typing.Callable
        The objective function we want to compute the signature of. This function must take inputs in the form of
        ``dict(str, object)`` with one dictionary key per variable, and provide `float` as the output.
    api_config : dict(str, dict)
        Configuration of the optimization variables. See API description.

    Returns
    -------
    signature_x : list(dict(str, object)) of shape (n_suggest,)
        The input locations probed on signature call.
    signature_y : list(float) of shape (n_suggest,)
        The objective function values at the inputs points. This is the real signature.
    """
    # Make sure get same sequence on every call to be a signature
    random = np.random.RandomState(0)

    signature_x = rs.suggest_dict([], [], api_config, n_suggestions=N_SUGGESTIONS, random=random)

    # For now, we only take the first output as the signature. We can generalize this later.
    signature_y = [f(xx)[0] for xx in signature_x]
    assert np.all(np.isfinite(signature_y)), "non-finite values found in signature for function"
    return signature_x, signature_y


def analyze_signatures(signatures):
    """Analyze function signatures from the experiment.

    Parameters
    ----------
    signatures : dict(str, list(list(float)))
        The signatures should all be the same length, so it should be 2D array
        like.

    Returns
    -------
    sig_errs : :class:`pandas:pandas.DataFrame`
        rows are test cases, columns are test points.
    signatures_median : dict(str, list(float))
        Median signature across all repetition per test case.
    """
    sig_errs = {}
    signatures_median = {}
    for test_case, signature_y in signatures.items():
        assert len(signature_y) > 0, "signature with no cases found"
        assert np.all(np.isfinite(signature_y)), "non-finite values found in signature for function"

        minval = np.min(signature_y, axis=0)
        maxval = np.max(signature_y, axis=0)

        if not np.allclose(minval, maxval):
            # Arguably, the util should not raise the warning, and these should
            # be raised on the outside, but let's do this for simplicity.
            warnings.warn(
                "Signature diverged on %s betwen %s and %s" % (test_case, str(minval), str(maxval)), RuntimeWarning
            )
        sig_errs[test_case] = maxval - minval
        # ensure serializable using tolist
        signatures_median[test_case] = np.median(signature_y, axis=0).tolist()

    # Convert to pandas so easy to append margins with max, better for disp.
    # If we let the user convert to pandas then we don't need dep on pandas.
    sig_errs = pd.DataFrame(sig_errs).T
    sig_errs.loc["max", :] = sig_errs.max(axis=0)
    sig_errs.loc[:, "max"] = sig_errs.max(axis=1)

    return sig_errs, signatures_median


def analyze_signature_pair(signatures, signatures_ref):
    """Analyze a pair of signatures (often from two sets of experiments) and return the error between them.

    Parameters
    ----------
    signatures : dict(str, list(float))
        Signatures from set of experiments. The signatures must all be the same length, so it should be 2D array like.
    signatures_ref : dict(str, list(float))
        The signatures from a reference set of experiments. The keys in `signatures` must be a subset of the signatures
        in `signatures_ref`.

    Returns
    -------
    sig_errs : :class:`pandas:pandas.DataFrame`
        rows are test cases, columns are test points.
    signatures_median : dict(str, list(float))
        Median signature across all repetition per test case.
    """
    signatures_pair = {kk: [signatures[kk], signatures_ref[kk]] for kk in signatures}
    sig_errs, signatures_pair = analyze_signatures(signatures_pair)
    return sig_errs, signatures_pair



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute quantiles and confidence intervals.
"""
import numpy as np
import scipy.stats as ss

from bayesmark.np_util import isclose_lte


def ensure_shape(x, y):
    """Util to broadcast on var to another but only when shape is different.

    This way we don't convert scalar into array type unnecessarily.
    """
    shape_y = np.shape(y)
    if np.shape(x) == shape_y:
        return x
    return np.broadcast_to(x, shape_y)


def order_stats(X):
    """Compute order statistics on sample `X`.

    Follows convention that order statistic 1 is minimum and statistic n is maximum. Therefore, array elements ``0``
    and ``n+1`` are ``-inf`` and ``+inf``.

    Parameters
    ----------
    X : :class:`numpy:numpy.ndarray` of shape (n,)
        Data for order statistics. Can be vectorized. Must be sortable data type (which is almost everything).

    Returns
    -------
    o_stats : :class:`numpy:numpy.ndarray` of shape (n+2,)
        Order statistics on `X`.
    """
    assert np.ndim(X) >= 1
    # NaN is not allowed since it does not have well defined order.
    assert not np.any(np.isnan(X))

    X_shape = np.shape(X)
    inf_pad = np.full(X_shape[:-1] + (1,), np.inf)

    o_stats = np.concatenate((-inf_pad, np.sort(X, axis=-1), inf_pad), axis=-1)
    return o_stats


def _quantile(n, q):
    idx = np.ceil(n * q).astype(int)
    return idx


def quantile(X, q):
    """Computes `q` th quantile of `X`.

    Similar to :func:`numpy:numpy.percentile` except that it matches the mathematical definition of a quantile *and*
    `q` is scaled in (0,1) rather than (0,100).

    Parameters
    ----------
    X : :class:`numpy:numpy.ndarray` of shape (n,)
        Data for quantile estimation. Can be vectorized. Must be sortable data type (which is almost everything).
    q : float
        Quantile to compute, must be in (0, 1). Can be vectorized.

    Returns
    -------
    estimate : dtype of `X`, scalar
        Empirical `q` quantile from sample `X`.
    """
    assert np.ndim(X) >= 1
    # We could robustify things to allow the edge cases, but maybe later
    assert np.all(0 < q) and np.all(q < 1)
    # Currently don't support broadcasting both at same time
    assert np.ndim(X) == 1 or np.ndim(q) == 0

    n = X.shape[-1]
    idx = _quantile(n, q)

    o_stats = order_stats(X)
    estimate = o_stats[..., idx]
    return estimate


def _quantile_CI(n, q, alpha):
    # Use in case there is -inf case from being at extreme of distn
    idx_lower = np.fmax(0, ss.binom.ppf(alpha / 2.0, n, q)).astype(int)
    assert np.all(isclose_lte(ss.binom.cdf(idx_lower - 1, n, q), alpha / 2.0))
    assert np.all(isclose_lte(alpha / 2.0, ss.binom.cdf(idx_lower, n, q)))
    assert np.all(0 <= idx_lower) and np.all(idx_lower <= n + 1)

    idx_upper = np.fmax(0, ss.binom.isf(alpha / 2.0, n, q)).astype(int) + 1
    assert np.all(isclose_lte(ss.binom.sf(idx_upper - 1, n, q), alpha / 2.0))
    assert np.all(isclose_lte(alpha / 2.0, ss.binom.sf(idx_upper - 2, n, q)))
    assert np.all(isclose_lte(1 - (alpha / 2.0), ss.binom.cdf(idx_upper - 1, n, q)))
    assert np.all(isclose_lte(ss.binom.cdf(idx_upper - 2, n, q), 1 - (alpha / 2.0)))
    assert np.all(0 <= idx_upper) and np.all(idx_upper <= n + 1)

    C = ss.binom.cdf(idx_upper - 1, n, q) - ss.binom.cdf(idx_lower - 1, n, q)
    assert np.all(isclose_lte(1.0 - alpha, C))

    return idx_lower, idx_upper


def quantile_CI(X, q, alpha=0.05):
    """Calculate CI on `q` quantile from same `X` using nonparametric estimation from order statistics.

    This will have alpha level of at most `alpha` due to the discrete nature of order statistics.

    Parameters
    ----------
    X : :class:`numpy:numpy.ndarray` of shape (n,)
        Data for quantile estimation. Can be vectorized. Must be sortable data type (which is almost everything).
    q : float
        Quantile to compute, must be in (0, 1). Can be vectorized.
    alpha : float
        False positive rate we allow for CI, must be in (0, 1). Can be vectorized.

    Returns
    -------
    LB : dtype of `X`, scalar
        Lower end on CI
    UB : dtype of `X`, scalar
        Upper end on CI
    """
    assert np.ndim(X) >= 1
    # We could robustify things to allow the edge cases, but maybe later
    assert np.all(0 < q) and np.all(q < 1)
    assert np.all(0 < alpha) and np.all(alpha < 1)
    # Currently don't support broadcasting both at same time
    assert np.ndim(X) == 1 or (np.ndim(q) == 0 and np.ndim(alpha) == 0)

    n = X.shape[-1]
    idx_lower, idx_upper = _quantile_CI(n, q, alpha)

    o_stats = order_stats(X)
    LB, UB = o_stats[..., idx_lower], o_stats[..., idx_upper]
    return LB, UB


def max_quantile_CI(X, q, m, alpha=0.05):
    """Calculate CI on `q` quantile of distribution on max of `m` iid samples using a data set `X`.

    This uses nonparametric estimation from order statistics and will have alpha level of at most `alpha` due to the
    discrete nature of order statistics.

    Parameters
    ----------
    X : :class:`numpy:numpy.ndarray` of shape (n,)
        Data for quantile estimation. Can be vectorized. Must be sortable data type (which is almost everything).
    q : float
        Quantile to compute, must be in (0, 1). Can be vectorized.
    m : int
        Compute statistics for distribution on max over `m` samples. Must be ``>= 1``. Can be vectorized.
    alpha : float
        False positive rate we allow for CI, must be in (0, 1). Can be vectorized.

    Returns
    -------
    estimate : dtype of `X`, scalar
        Best estimate on `q` quantile on max over `m` iid samples.
    LB : dtype of `X`, scalar
        Lower end on CI
    UB : dtype of `X`, scalar
        Upper end on CI
    """
    # X and alpha used/checked below in quantile_CI routine.
    # We could robustify things to allow the edge cases, but maybe later
    assert np.all(0 < q) and np.all(q < 1)
    # Could check int but if someone wants to interpolate, we will let them.
    assert np.all(m >= 1)
    # Currently don't support broadcasting both at same time
    assert np.ndim(X) == 1 or (np.ndim(q) == 0 and np.ndim(q) == 0 and np.ndim(alpha) == 0)

    q = q ** (1.0 / m)
    o_stats = order_stats(X)

    n = X.shape[-1]
    idx = _quantile(n, q)
    idx_lower, idx_upper = _quantile_CI(n, q, alpha=alpha)

    LB, UB = o_stats[..., idx_lower], o_stats[..., idx_upper]
    # Might need to broadcast estimate out if vectorization is in alpha
    estimate = ensure_shape(o_stats[..., idx], LB)
    return estimate, LB, UB


def min_quantile_CI(X, q, m, alpha=0.05):
    """Calculate confidence interval on `q` quantile of distribution on min of `m` iid samples using a data set `X`.

    This uses nonparametric estimation from order statistics and will have alpha level of at most `alpha` due to the
    discrete nature of order statistics.

    Parameters
    ----------
    X : :class:`numpy:numpy.ndarray` of shape (n,)
        Data for quantile estimation. Can be vectorized. Must be sortable data type (which is almost everything).
    q : float
        Quantile to compute, must be in (0, 1). Can be vectorized.
    m : int
        Compute statistics for distribution on min over `m` samples. Must be ``>= 1``. Can be vectorized.
    alpha : float
        False positive rate we allow for CI, must be in (0, 1). Can be vectorized.

    Returns
    -------
    estimate : dtype of `X`, scalar
        Best estimate on `q` quantile on min over `m` iid samples.
    LB : dtype of `X`, scalar
        Lower end on CI
    UB : dtype of `X`, scalar
        Upper end on CI
    """
    # X and alpha used/checked below in quantile_CI routine.
    # We could robustify things to allow the edge cases, but maybe later
    assert np.all(0 < q) and np.all(q < 1)
    # Could check int but if someone wants to interp, we will let them.
    assert np.all(m >= 1)
    # Currently don't support broadcasting both at same time
    assert np.ndim(X) == 1 or (np.ndim(q) == 0 and np.ndim(q) == 0 and np.ndim(alpha) == 0)

    # This might have numerics issues for small q
    q = 1.0 - (1.0 - q) ** (1.0 / m)
    o_stats = order_stats(X)

    n = X.shape[-1]
    idx = _quantile(n, q)
    idx_lower, idx_upper = _quantile_CI(n, q, alpha=alpha)

    LB, UB = o_stats[..., idx_lower], o_stats[..., idx_upper]
    # Might need to broadcast estimate out if vectorization is in alpha
    estimate = ensure_shape(o_stats[..., idx], LB)
    return estimate, LB, UB


def quantile_and_CI(X, q, alpha=0.05):
    """Calculate CI on `q` quantile from same `X` using nonparametric estimation from order statistics.

    This will have alpha level of at most `alpha` due to the discrete nature of order statistics.

    Parameters
    ----------
    X : :class:`numpy:numpy.ndarray` of shape (n,)
        Data for quantile estimation. Can be vectorized. Must be sortable data type (which is almost everything).
    q : float
        Quantile to compute, must be in (0, 1). Can be vectorized.
    alpha : float
        False positive rate we allow for CI, must be in (0, 1). Can be vectorized.

    Returns
    -------
    estimate : dtype of `X`, scalar
        Empirical `q` quantile from sample `X`.
    LB : dtype of `X`, scalar
        Lower end on CI
    UB : dtype of `X`, scalar
        Upper end on CI
    """
    # This routine is mostly just a wrapper routine
    estimate, LB, UB = max_quantile_CI(X, q=q, m=1, alpha=alpha)
    return estimate, LB, UB



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract base class for the optimizers in the benchmark. This creates a common API across all packages.
"""
from abc import ABC, abstractmethod

from importlib_metadata import version


class AbstractOptimizer(ABC):
    """Abstract base class for the optimizers in the benchmark. This creates a common API across all packages.
    """

    # Every implementation package needs to specify this static variable, e.g., "primary_import=opentuner"
    primary_import = None

    def __init__(self, api_config, **kwargs):
        """Build wrapper class to use an optimizer in benchmark.

        Parameters
        ----------
        api_config : dict-like of dict-like
            Configuration of the optimization variables. See API description.
        """
        self.api_config = api_config

    @classmethod
    def get_version(cls):
        """Get the version for this optimizer.

        Returns
        -------
        version_str : str
            Version number of the optimizer. Usually, this is equivalent to ``package.__version__``.
        """
        assert (cls.primary_import is None) or isinstance(cls.primary_import, str)
        # Should use x.x.x as version if sub-class did not specify its primary import
        version_str = "x.x.x" if cls.primary_import is None else version(cls.primary_import)
        return version_str

    @abstractmethod
    def suggest(self, n_suggestions):
        """Get a suggestion from the optimizer.

        Parameters
        ----------
        n_suggestions : int
            Desired number of parallel suggestions in the output

        Returns
        -------
        next_guess : list of dict
            List of `n_suggestions` suggestions to evaluate the objective
            function. Each suggestion is a dictionary where each key
            corresponds to a parameter being optimized.
        """
        pass

    @abstractmethod
    def observe(self, X, y):
        """Send an observation of a suggestion back to the optimizer.

        Parameters
        ----------
        X : list of dict-like
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,)
            Corresponding values where objective has been evaluated
        """
        pass



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Aggregate the results of many studies to prepare analysis.
"""
import json
import logging
from collections import Counter

import numpy as np
import xarray as xr

import bayesmark.constants as cc
import bayesmark.xr_util as xru
from bayesmark.cmd_parse import CmdArgs, agg_parser, parse_args, serializable_dict, unserializable_dict
from bayesmark.constants import ARG_DELIM, EVAL_RESULTS, ITER, METHOD, SUGGEST, TEST_CASE, TIME_RESULTS, TRIAL
from bayesmark.serialize import XRSerializer
from bayesmark.signatures import analyze_signatures
from bayesmark.sklearn_funcs import SklearnModel
from bayesmark.util import str_join_safe

logger = logging.getLogger(__name__)


def validate_time(all_time):
    """Validate the aggregated time data set."""
    assert isinstance(all_time, xr.Dataset)
    assert all_time[cc.SUGGEST_PHASE].dims == (ITER,)
    assert all_time[cc.EVAL_PHASE].dims == (ITER, SUGGEST)
    assert all_time[cc.OBS_PHASE].dims == (ITER,)
    assert xru.is_simple_coords(all_time.coords, min_side=1)


def validate_perf(perf_da):
    """Validate the input eval data arrays."""
    assert isinstance(perf_da, xr.Dataset)
    assert perf_da.dims == (ITER, SUGGEST)
    assert xru.is_simple_coords(perf_da.coords)
    assert not np.any(np.isnan(perf_da.values))


def validate_agg_perf(perf_da, min_trial=1):
    """Validate the aggregated eval data set."""
    assert isinstance(perf_da, xr.DataArray)
    assert perf_da.dims == (ITER, SUGGEST, TEST_CASE, METHOD, TRIAL)
    assert xru.is_simple_coords(perf_da.coords, dims=(ITER, SUGGEST, TRIAL))
    assert not np.any(np.isnan(perf_da.values))
    assert perf_da.sizes[TRIAL] >= min_trial


def summarize_time(all_time):
    """Transform a single timing dataset from an experiment into a form better for aggregation.

    Parameters
    ----------
    all_time : :class:`xarray:xarray.Dataset`
        Dataset with variables ``(SUGGEST_PHASE, EVAL_PHASE, OBS_PHASE)`` which have dimensions ``(ITER,)``,
        ``(ITER, SUGGEST)``, and ``(ITER,)``, respectively. The variable `EVAL_PHASE` has the function evaluation time
        for each parallel suggestion.

    Returns
    -------
    time_summary : :class:`xarray:xarray.Dataset`
        Dataset with variables ``(SUGGEST_PHASE, OBS_PHASE, EVAL_PHASE_MAX, EVAL_PHASE_SUM)`` which all have dimensions
        ``(ITER,)``. The maximum `EVAL_PHASE_MAX` is relevant for wall clock time, while `EVAL_PHASE_SUM` is relevant
        for CPU time.
    """
    validate_time(all_time)

    time_summary = xr.Dataset(coords=all_time.coords)

    time_summary[cc.SUGGEST_PHASE] = all_time[cc.SUGGEST_PHASE]
    time_summary[cc.OBS_PHASE] = all_time[cc.OBS_PHASE]
    time_summary[cc.EVAL_PHASE_MAX] = all_time[cc.EVAL_PHASE].max(dim=SUGGEST)
    time_summary[cc.EVAL_PHASE_SUM] = all_time[cc.EVAL_PHASE].sum(dim=SUGGEST)
    return time_summary


def concat_experiments(all_experiments, ravel=False):
    """Aggregate the Datasets from a series of experiments into combined Dataset.

    Parameters
    ----------
    all_experiments : typing.Iterable
        Iterable (possible from a generator) with the Datasets from each experiment. Each item in `all_experiments` is
        a pair containing ``(meta_data, data)``. See `load_experiments` for details on these variables,
    ravel : bool
        If true, ravel all studies to store batch suggestions as if they were serial.

    Returns
    -------
    all_perf : :class:`xarray:xarray.Dataset`
        DataArray containing all of the `perf_da` from the experiments. The meta-data from the experiments are included
        as extra dimensions. `all_perf` has dimensions ``(ITER, SUGGEST, TEST_CASE, METHOD, TRIAL)``. To convert the
        `uuid` to a trial, there must be an equal number of repetition in the experiments for each `TEST_CASE`,
        `METHOD` combination. Likewise, all of the experiments need an equal number of `ITER` and `SUGGEST`. If `ravel`
        is true, then the `SUGGEST` is singleton.
    all_time : :class:`xarray:xarray.Dataset`
        Dataset containing all of the `time_ds` from the experiments. The new dimensions are
        ``(ITER, TEST_CASE, METHOD, TRIAL)``. It has the same variables as `time_ds`.
    all_suggest : :class:`xarray:xarray.Dataset`
        DataArray containing all of the `suggest_ds` from the experiments. It has dimensions
        ``(ITER, SUGGEST, TEST_CASE, METHOD, TRIAL)``.
    all_sigs : dict(str, list(list(float)))
        Aggregate of all experiment signatures.
    """
    all_perf = {}
    all_time = {}
    all_suggest = {}
    all_sigs = {}
    trial_counter = Counter()
    for (test_case, optimizer, uuid), (perf_ds, time_ds, suggest_ds, sig) in all_experiments:
        if ravel:
            raise NotImplementedError("ravel is deprecated. Just reshape in analysis steps instead.")

        case_key = (test_case, optimizer, trial_counter[(test_case, optimizer)])
        trial_counter[(test_case, optimizer)] += 1

        # Process perf data
        assert all(perf_ds[kk].dims == (ITER, SUGGEST) for kk in perf_ds)
        all_perf[case_key] = perf_ds

        # Process time data
        all_time[case_key] = summarize_time(time_ds)

        # Process suggestion data
        all_suggest_curr = all_suggest.setdefault(test_case, {})
        all_suggest_curr[case_key] = suggest_ds

        # Handle the signatures
        all_sigs.setdefault(test_case, []).append(sig)
    assert min(trial_counter.values()) == max(trial_counter.values()), "Uneven number of trials per test case"

    # Now need to concat dict of datasets into single dataset
    all_perf = xru.ds_concat(all_perf, dims=(TEST_CASE, METHOD, TRIAL))
    assert all(all_perf[kk].dims == (ITER, SUGGEST, TEST_CASE, METHOD, TRIAL) for kk in all_perf)
    assert not any(
        np.any(np.isnan(all_perf[kk].values)) for kk in all_perf
    ), "Missing combinations of method and test case"

    all_time = xru.ds_concat(all_time, dims=(TEST_CASE, METHOD, TRIAL))
    assert all(all_time[kk].dims == (ITER, TEST_CASE, METHOD, TRIAL) for kk in all_time)
    assert not any(np.any(np.isnan(all_time[kk].values)) for kk in all_time)
    assert xru.coord_compat((all_perf, all_time), (ITER, TEST_CASE, METHOD, TRIAL))

    for test_case in all_suggest:
        all_suggest[test_case] = xru.ds_concat(all_suggest[test_case], dims=(TEST_CASE, METHOD, TRIAL))
        assert all(
            all_suggest[test_case][kk].dims == (ITER, SUGGEST, TEST_CASE, METHOD, TRIAL)
            for kk in all_suggest[test_case]
        )
        assert not any(np.any(np.isnan(all_suggest[test_case][kk].values)) for kk in all_suggest[test_case])
        assert xru.coord_compat((all_perf, all_suggest[test_case]), (ITER, METHOD, TRIAL))
        assert all_suggest[test_case].coords[TEST_CASE].shape == (1,), "test case should be singleton"

    return all_perf, all_time, all_suggest, all_sigs


def load_experiments(uuid_list, db_root, dbid):  # pragma: io
    """Generator to load the results of the experiments.

    Parameters
    ----------
    uuid_list : list(uuid.UUID)
        List of UUIDs corresponding to experiments to load.
    db_root : str
        Root location for data store as requested by the serializer used.
    dbid : str
        Name of the data store as requested by the serializer used.

    Yields
    ------
    meta_data : (str, str, str)
        The `meta_data` contains a `tuple` of `str` with ``test_case, optimizer, uuid``.
    data : (:class:`xarray:xarray.Dataset`, :class:`xarray:xarray.Dataset`, :class:`xarray:xarray.Dataset` list(float))
        The `data` contains a tuple of ``(perf_ds, time_ds, suggest_ds, sig)``. The `perf_ds` is a
        :class:`xarray:xarray.Dataset` containing the evaluation results with dimensions ``(ITER, SUGGEST)``, each
        variable is an objective. The `time_ds` is an :class:`xarray:xarray.Dataset` containing the timing results of
        the form accepted by `summarize_time`. The coordinates must be compatible with `perf_ds`. The suggest_ds is a
        :class:`xarray:xarray.Dataset` containing the inputs to the function evaluations. Each variable is a function
        input. Finally, `sig` contains the `test_case` signature and must be `list(float)`.
    """
    uuids_seen = set()
    for uuid_ in uuid_list:
        logger.info(uuid_.hex)

        # Load perf and timing data
        perf_ds, meta = XRSerializer.load(db_root, db=dbid, key=cc.EVAL, uuid_=uuid_)
        time_ds, meta_t = XRSerializer.load(db_root, db=dbid, key=cc.TIME, uuid_=uuid_)
        assert meta == meta_t, "meta data should between time and eval files"
        suggest_ds, meta_t = XRSerializer.load(db_root, db=dbid, key=cc.SUGGEST_LOG, uuid_=uuid_)
        assert meta == meta_t, "meta data should between suggest and eval files"

        # Get signature to pass out as well
        _, sig = meta["signature"]
        logger.info(meta)
        logger.info(sig)

        # Build the new indices for combined data, this could be put in function for easier testing
        eval_args = unserializable_dict(meta["args"])  # Unpack meta-data
        test_case = SklearnModel.test_case_str(
            eval_args[CmdArgs.classifier], eval_args[CmdArgs.data], eval_args[CmdArgs.metric]
        )
        optimizer = str_join_safe(
            ARG_DELIM, (eval_args[CmdArgs.optimizer], eval_args[CmdArgs.opt_rev], eval_args[CmdArgs.rev])
        )
        args_uuid = eval_args[CmdArgs.uuid]

        # Check UUID sanity
        assert isinstance(args_uuid, str)
        assert args_uuid == uuid_.hex, "UUID meta-data does not match filename"
        assert args_uuid not in uuids_seen, "uuids being reused between studies"
        uuids_seen.add(args_uuid)

        # Return key -> data so this generator can be iterated over in dict like manner
        meta_data = (test_case, optimizer, args_uuid)
        data = (perf_ds, time_ds, suggest_ds, sig)
        yield meta_data, data


def main():
    """See README for instructions on calling aggregate.
    """
    description = "Aggregate study results across functions and optimizers"
    args = parse_args(agg_parser(description))

    logger.setLevel(logging.INFO)  # Note this is the module-wide logger
    if args[CmdArgs.verbose]:
        logger.addHandler(logging.StreamHandler())

    # Get list of UUIDs
    uuid_list = XRSerializer.get_uuids(args[CmdArgs.db_root], db=args[CmdArgs.db], key=cc.EVAL)
    uuid_list_ = XRSerializer.get_uuids(args[CmdArgs.db_root], db=args[CmdArgs.db], key=cc.TIME)
    assert uuid_list == uuid_list_, "UUID list does not match between time and eval results"
    uuid_list_ = XRSerializer.get_uuids(args[CmdArgs.db_root], db=args[CmdArgs.db], key=cc.SUGGEST_LOG)
    assert uuid_list == uuid_list_, "UUID list does not match between suggest log and eval results"

    # Get iterator of all experiment data dumps, load in and process, and concat
    data_G = load_experiments(uuid_list, args[CmdArgs.db_root], args[CmdArgs.db])
    all_perf, all_time, all_suggest, all_sigs = concat_experiments(data_G, ravel=args[CmdArgs.ravel])

    # Check the concat signatures make are coherent
    sig_errs, signatures_median = analyze_signatures(all_sigs)
    logger.info("Signature errors:\n%s" % sig_errs.to_string())
    print(json.dumps({"exp-agg sig errors": sig_errs.T.to_dict()}))

    # Dump and save it all out
    logger.info("saving")
    meta = {"args": serializable_dict(args), "signature": signatures_median}
    XRSerializer.save_derived(all_perf, meta, args[CmdArgs.db_root], db=args[CmdArgs.db], key=EVAL_RESULTS)
    XRSerializer.save_derived(all_time, meta, args[CmdArgs.db_root], db=args[CmdArgs.db], key=TIME_RESULTS)
    for test_case, ds in all_suggest.items():
        XRSerializer.save_derived(ds, meta, args[CmdArgs.db_root], db=args[CmdArgs.db], key=test_case)

    logger.info("done")


if __name__ == "__main__":
    main()  # pragma: main



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Anything related to parsing command line arguments goes in here. There are some custom data structures to
represent all the options available in the experiments here.

Not currently any doc strings in this file because it may become obsolete with the use of fire package.
"""
import argparse
import json
import os.path
import sys
import uuid as pyuuid
from enum import IntEnum, auto
from pathlib import PosixPath

import git
from git.exc import InvalidGitRepositoryError
from pathvalidate.argparse import sanitize_filename, validate_filename, validate_filepath

from bayesmark.builtin_opt.config import CONFIG
from bayesmark.constants import ARG_DELIM, DATA_LOADER_NAMES, METRICS, MODEL_NAMES, OPTIMIZERS_FILE, PY_INTERPRETER
from bayesmark.path_util import absopen, abspath
from bayesmark.util import shell_join

assert not any(ARG_DELIM in opt for opt in MODEL_NAMES)
assert not any(ARG_DELIM in opt for opt in DATA_LOADER_NAMES)


class CmdArgs(IntEnum):
    uuid = auto()
    db_root = auto()
    optimizer_root = auto()
    data_root = auto()
    db = auto()
    optimizer = auto()
    data = auto()
    classifier = auto()
    metric = auto()
    n_calls = auto()
    n_suggest = auto()
    n_repeat = auto()
    n_jobs = auto()
    jobs_file = auto()
    ravel = auto()
    verbose = auto()
    dry_run = auto()
    rev = auto()
    opt_rev = auto()


CMD_STR = {
    CmdArgs.uuid: ("-u", "--uuid"),
    CmdArgs.db_root: ("-dir", "-db-root"),
    CmdArgs.optimizer_root: ("-odir", "--opt-root"),
    CmdArgs.data_root: ("-dr", "--data-root"),
    CmdArgs.db: ("-b", "--db"),
    CmdArgs.optimizer: ("-o", "--opt"),
    CmdArgs.data: ("-d", "--data"),
    CmdArgs.classifier: ("-c", "--classifier"),
    CmdArgs.metric: ("-m", "--metric"),
    CmdArgs.n_calls: ("-n", "--calls"),
    CmdArgs.n_suggest: ("-p", "--suggestions"),
    CmdArgs.n_repeat: ("-r", "--repeat"),
    CmdArgs.n_jobs: ("-nj", "--num-jobs"),
    CmdArgs.jobs_file: ("-ofile", "--jobs-file"),
    CmdArgs.ravel: ("-rv", "--ravel"),
    CmdArgs.verbose: ("-v", "--verbose"),
    CmdArgs.dry_run: (None, "dry_run"),  # Will not be specified from CLI
    CmdArgs.rev: (None, "rev"),  # Will not be specified from CLI
    CmdArgs.opt_rev: (None, "opt_rev"),  # Will not be specified from CLI. Which version of optimizer.
}


def arg_to_str(arg):
    # We can change this so it is arg.value, or someway to be usable by field interface
    _, dest = str(arg).split(".")
    return dest


def namespace_to_dict(args_ns):
    args = vars(args_ns)
    args = {kk: args[arg_to_str(kk)] for kk in CMD_STR if (arg_to_str(kk) in args)}
    return args


def serializable_dict(args):
    args_str = {CMD_STR[kk][1]: args[kk] for kk in CMD_STR if (kk in args)}
    assert len(args_str) == len(args)
    return args_str


def unserializable_dict(args_str):
    args = {kk: args_str[CMD_STR[kk][1]] for kk in CMD_STR if (CMD_STR[kk][1] in args_str)}
    assert len(args_str) == len(args)
    return args


def add_argument(parser, arg, **kwargs):
    short_name, long_name = CMD_STR[arg]
    dest = arg_to_str(arg)
    parser.add_argument(short_name, long_name, dest=dest, **kwargs)


def filepath(value):
    """Work around for `pathvalidate` bug."""
    if value == ".":
        return value
    validate_filepath(value, platform="auto")
    return value


def filename(value):
    validate_filename(value, platform="universal")
    return value


def uuid(val_str):
    val = str.lower(val_str)
    uuid_ = pyuuid.UUID(hex=val)
    assert val == uuid_.hex, "error in parsing uuid"
    return val


def positive_int(val_str):
    val = int(val_str)
    if val <= 0:
        msg = "expected positive, got %s" % val_str
        raise argparse.ArgumentTypeError(msg)
    return val


def joinable(val_str):
    val = str(val_str)  # just for good measure
    validate_filename(val, platform="universal")  # we choose to be at least as strict as filenames
    if ARG_DELIM in val:
        msg = "delimiter %s not allowed in choice %s" % (ARG_DELIM, val)
        raise argparse.ArgumentTypeError(msg)
    return val


def load_rev_number():
    # This function uses a lot of language "power features" that could be considered bad form:
    # 1) does a conditional import to get version
    # 2) uses __file__ to try and extract and git repo version during execution
    # We will let this fly anyway because:
    # 1) The results of this are only used for logging anyway
    # 2) This is a command parsing module of the code and inherently very non-pure and doing IO etc
    # 3) Unclear if there is a cleaner way to do this

    # Get rev from version file (if running inside the pip-installable wheel without the git repo)
    try:
        from bayesmark import version

        rev_file = version.VERSION
    except ImportError:
        rev_file = None
    else:
        rev_file = rev_file.strip()
        rev = rev_file

    # Get rev from git API if inside git repo (and not built wheel from pip install ...)
    wdir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
    try:
        repo = git.Repo(path=wdir, search_parent_directories=False)
    except InvalidGitRepositoryError:
        rev_repo = None
    else:
        rev_repo = repo.head.commit.hexsha
        rev_repo = rev_repo.strip()
        rev = rev_repo

    # Check coherence of what we found
    if (rev_repo is None) and (rev_file is None):
        raise RuntimeError("Must specify version.py if not inside a git repo.")
    if (rev_repo is not None) and (rev_file is not None):
        assert rev_repo == rev_file, "Rev file %s does not match rev git %s" % (rev_file, rev_repo)

    assert rev == rev.strip()
    # We could first enforce is_lower_hex if we want to enforce that
    rev = rev[:7]
    return rev


def base_parser():
    parser = argparse.ArgumentParser(add_help=False)

    add_argument(
        parser, CmdArgs.db_root, default=".", type=filepath, help="root directory for all benchmark experiments output"
    )
    add_argument(
        parser, CmdArgs.optimizer_root, default=".", type=filepath, help="Directory with optimization wrappers"
    )
    # Always a verbose flag option
    add_argument(parser, CmdArgs.verbose, action="store_true", help="print the study logs to console")
    return parser


def launcher_parser(description):
    parser = argparse.ArgumentParser(description=description, parents=[base_parser()])

    add_argument(parser, CmdArgs.uuid, type=uuid, help="length 32 hex UUID for this experiment")
    add_argument(parser, CmdArgs.data_root, type=filepath, help="root directory for all custom csv files")
    add_argument(parser, CmdArgs.db, type=filename, help="database ID of this benchmark experiment")

    add_argument(parser, CmdArgs.optimizer, type=joinable, nargs="+", help="optimizers to use")
    add_argument(parser, CmdArgs.data, type=joinable, nargs="+", help="data sets to use")
    add_argument(parser, CmdArgs.classifier, type=joinable, nargs="+", help="classifiers to use")
    add_argument(parser, CmdArgs.metric, type=str, choices=METRICS, nargs="+", help="scoring metric to use")

    # Iterations counts used in experiments
    add_argument(parser, CmdArgs.n_calls, default=100, type=positive_int, help="number of function evaluations")
    add_argument(
        parser, CmdArgs.n_suggest, default=1, type=positive_int, help="number of suggestions to provide in parallel"
    )
    add_argument(parser, CmdArgs.n_repeat, default=20, type=positive_int, help="number of repetitions of each study")

    # Arguments for creating dry run jobs file
    add_argument(
        parser,
        CmdArgs.n_jobs,
        type=int,
        default=0,
        help="number of jobs to put in the dry run file, the default 0 value disables dry run (real run)",
    )
    # Using default of current dir for jobs file output since that is generally the default for everything
    add_argument(
        parser, CmdArgs.jobs_file, type=filepath, default="./jobs.txt", help="a jobs file with all commands to be run"
    )
    return parser


def experiment_parser(description):
    parser = argparse.ArgumentParser(description=description, parents=[base_parser()])

    add_argument(parser, CmdArgs.uuid, type=uuid, required=True, help="length 32 hex UUID for this experiment")

    # This could be made simpler and use '.' default for dataroot, even if no custom data used.
    add_argument(parser, CmdArgs.data_root, type=filepath, help="root directory for all custom csv files")
    add_argument(parser, CmdArgs.db, type=filename, required=True, help="database ID of this benchmark experiment")
    add_argument(parser, CmdArgs.optimizer, required=True, type=joinable, help="optimizer to use")
    add_argument(parser, CmdArgs.data, required=True, type=joinable, help="data set to use")
    add_argument(parser, CmdArgs.classifier, required=True, type=joinable, help="classifier to use")
    add_argument(parser, CmdArgs.metric, required=True, type=str, choices=METRICS, help="scoring metric to use")

    add_argument(parser, CmdArgs.n_calls, default=100, type=positive_int, help="number of function evaluations")
    add_argument(
        parser, CmdArgs.n_suggest, default=1, type=positive_int, help="number of suggestions to provide in parallel"
    )
    return parser


def agg_parser(description):
    parser = argparse.ArgumentParser(description=description, parents=[base_parser()])
    add_argument(parser, CmdArgs.db, type=filename, required=True, help="database ID of this benchmark experiment")
    add_argument(
        parser,
        CmdArgs.ravel,
        action="store_true",
        help="ravel all studies to store batch suggestions as if they were serial (deprecated)",
    )
    return parser


def general_parser(description):
    parser = argparse.ArgumentParser(description=description, parents=[base_parser()])
    add_argument(parser, CmdArgs.db, type=filename, required=True, help="database ID of this benchmark experiment")
    return parser


def parse_args(parser, argv=None):
    """Note that this argument parser does not check compatibility between clf/reg metric and data set.
    """
    args = parser.parse_args(argv)
    args = namespace_to_dict(args)

    args[CmdArgs.dry_run] = (CmdArgs.n_jobs in args) and (args[CmdArgs.n_jobs] > 0)
    # Does not check dir actually exists here, but whatever
    args[CmdArgs.jobs_file] = abspath(args[CmdArgs.jobs_file], verify=False) if args[CmdArgs.dry_run] else None

    # Then make sure all path vars are abspath:
    # Dry run might be executing on diff system => cannot verify yet
    args[CmdArgs.db_root] = abspath(args[CmdArgs.db_root], verify=not args[CmdArgs.dry_run])
    args[CmdArgs.optimizer_root] = abspath(args[CmdArgs.optimizer_root], verify=True)
    if (CmdArgs.data_root in args) and (args[CmdArgs.data_root] is not None):
        args[CmdArgs.data_root] = abspath(args[CmdArgs.data_root], verify=not args[CmdArgs.dry_run])

    # Get git version of the benchmark itself for meta-data, just in case we need it.
    args[CmdArgs.rev] = load_rev_number()

    # We may support ability to specify version at args in the future, from now it is implied
    args[CmdArgs.opt_rev] = None
    return args


def _cleanup(filename_str):
    filename_str = sanitize_filename(filename_str, replacement_text="-", platform="universal")
    filename_str = filename_str.replace(ARG_DELIM, "-")
    return filename_str


def infer_settings(opt_root, opt_pattern="**/optimizer.py"):
    opt_root = PosixPath(opt_root)

    source_files = sorted(opt_root.glob(opt_pattern))
    source_files = [ss.relative_to(opt_root) for ss in source_files]

    settings = {_cleanup(str(ss.parent)): [str(ss), {}] for ss in source_files}

    assert all(joinable(kk) for kk in settings), "Something went wrong in name sanitization."
    assert len(settings) == len(source_files), "Name collision after sanitization of %s" % repr(source_files)
    assert len(set(CONFIG.keys()) & set(settings.keys())) == 0, "Name collision with builtin optimizers."

    return settings


def load_optimizer_settings(opt_root):
    try:
        with absopen(os.path.join(opt_root, OPTIMIZERS_FILE), "r") as f:
            settings = json.load(f)
    except FileNotFoundError:
        # Search for optimizers instead
        settings = infer_settings(opt_root)

    assert isinstance(settings, dict)
    assert not any((ARG_DELIM in opt) for opt in settings), "optimizer names violates name convention"
    return settings


def cmd_str():
    cmd = "%s %s" % (PY_INTERPRETER, shell_join(sys.argv))
    return cmd



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Perform analysis to compare different optimizers across problems.
"""
import json
import logging
import warnings
from collections import OrderedDict

import numpy as np
import xarray as xr

import bayesmark.constants as cc
import bayesmark.quantiles as qt
import bayesmark.xr_util as xru
from bayesmark.cmd_parse import CmdArgs, general_parser, parse_args, serializable_dict
from bayesmark.constants import (
    ITER,
    LB_MEAN,
    LB_MED,
    LB_NORMED_MEAN,
    METHOD,
    NORMED_MEAN,
    NORMED_MED,
    OBJECTIVE,
    PERF_BEST,
    PERF_CLIP,
    PERF_MEAN,
    PERF_MED,
    SUGGEST,
    TEST_CASE,
    TRIAL,
    UB_MEAN,
    UB_MED,
    UB_NORMED_MEAN,
)
from bayesmark.experiment_aggregate import validate_agg_perf
from bayesmark.experiment_baseline import do_baseline
from bayesmark.np_util import linear_rescale
from bayesmark.serialize import XRSerializer
from bayesmark.signatures import analyze_signature_pair
from bayesmark.stats import t_EB

# Mathematical settings
EVAL_Q = 0.5  # Evaluate based on median loss across n_trials
ALPHA = 0.05  # ==> 95% CIs

logger = logging.getLogger(__name__)


def cummin(x_val, x_key):
    """Get the cumulative minimum of `x_val` when ranked according to `x_key`.

    Parameters
    ----------
    x_val : :class:`numpy:numpy.ndarray` of shape (n, d)
        The array to get the cumulative minimum of along axis 0.
    x_key : :class:`numpy:numpy.ndarray` of shape (n, d)
        The array for ranking elements as to what is the minimum.

    Returns
    -------
    c_min : :class:`numpy:numpy.ndarray` of shape (n, d)
        The cumulative minimum array.
    """
    assert x_val.shape == x_key.shape
    assert x_val.ndim == 2
    n, _ = x_val.shape

    xm = np.minimum.accumulate(x_key, axis=0)
    idx = np.maximum.accumulate((x_key <= xm) * np.arange(n)[:, None])
    c_min = np.take_along_axis(x_val, idx, axis=0)
    return c_min


def get_perf_array(evals, evals_visible):
    """Get the actual (e.g., generalization loss) over iterations.

    Parameters
    ----------
    evals : :class:`numpy:numpy.ndarray` of shape (n_iter, n_batch, n_trials)
        The actual loss (e.g., generalization) for a given experiment.
    evals_visible : :class:`numpy:numpy.ndarray` of shape (n_iter, n_batch, n_trials)
        The observable loss (e.g., validation) for a given experiment.

    Returns
    -------
    perf_array : :class:`numpy:numpy.ndarray` of shape (n_iter, n_trials)
        The best performance so far at iteration i from `evals`. Where the best has been selected according to
        `evals_visible`.
    """
    n_iter, _, n_trials = evals.shape

    idx = np.argmin(evals_visible, axis=1)
    perf_array = np.take_along_axis(evals, idx[:, None, :], axis=1).squeeze(axis=1)
    assert perf_array.shape == (n_iter, n_trials)

    visible_perf_array = np.min(evals_visible, axis=1)
    assert visible_perf_array.shape == (n_iter, n_trials)

    # Get the minimum from the visible loss
    perf_array = cummin(perf_array, visible_perf_array)
    return perf_array


def compute_aggregates(perf_da, baseline_ds, visible_perf_da=None):
    """Aggregate function evaluations in the experiments to get performance summaries of each method.

    Parameters
    ----------
    perf_da : :class:`xarray:xarray.DataArray`
        Aggregate experimental results with each function evaluation in the experiments according to true loss
        (e.g., generalization). `perf_da` has dimensions ``(ITER, SUGGEST, TEST_CASE, METHOD, TRIAL)`` as is assumed
        to have no nan values.
    baseline_ds : :class:`xarray:xarray.Dataset`
        Dataset with baseline performance. It was variables ``(PERF_MED, PERF_MEAN, PERF_CLIP, PERF_BEST)`` with
        dimensions ``(ITER, TEST_CASE)``, ``(ITER, TEST_CASE)``, ``(TEST_CASE,)``, and ``(TEST_CASE,)``, respectively.
        `PERF_MED` is a baseline of performance based on random search when using medians to summarize performance.
        Likewise, `PERF_MEAN` is for means. `PERF_CLIP` is an upperbound to clip poor performance when using the mean.
        `PERF_BEST` is an estimate on the global minimum.
    visible_perf_da : :class:`xarray:xarray.DataArray`
        Aggregate experimental results with each function evaluation in the experiments according to visible loss
        (e.g., validation). `visible_perf_da` has dimensions ``(ITER, SUGGEST, TEST_CASE, METHOD, TRIAL)`` as is
        assumed to have no nan values. If `None`, we set ``visible_perf_da = perf_da``.

    Returns
    -------
    agg_result : :class:`xarray:xarray.Dataset`
        Dataset with summary of performance for each method and test case combination. Contains variables:
        ``(PERF_MED, LB_MED, UB_MED, NORMED_MED, PERF_MEAN, LB_MEAN, UB_MEAN, NORMED_MEAN)``
        each with dimensions ``(ITER, METHOD, TEST_CASE)``. `PERF_MED` is a median summary of performance with `LB_MED`
        and `UB_MED` as error bars. `NORMED_MED` is a rescaled `PERF_MED` so we expect the optimal performance is 0,
        and random search gives 1 at all `ITER`. Likewise, `PERF_MEAN`, `LB_MEAN`, `UB_MEAN`, `NORMED_MEAN` are for
        mean performance.
    summary : :class:`xarray:xarray.Dataset`
        Dataset with overall summary of performance of each method. Contains variables
        ``(PERF_MED, LB_MED, UB_MED, PERF_MEAN, LB_MEAN, UB_MEAN)``
        each with dimensions ``(ITER, METHOD)``.
    """
    validate_agg_perf(perf_da, min_trial=1)

    assert isinstance(baseline_ds, xr.Dataset)
    assert tuple(baseline_ds[PERF_BEST].dims) == (TEST_CASE,)
    assert tuple(baseline_ds[PERF_CLIP].dims) == (TEST_CASE,)
    assert tuple(baseline_ds[PERF_MED].dims) == (ITER, TEST_CASE)
    assert tuple(baseline_ds[PERF_MEAN].dims) == (ITER, TEST_CASE)
    assert xru.coord_compat((perf_da, baseline_ds), (ITER, TEST_CASE))
    assert not any(np.any(np.isnan(baseline_ds[kk].values)) for kk in baseline_ds)

    # Now actually get the aggregate performance numbers per test case
    agg_result = xru.ds_like(
        perf_da,
        (PERF_MED, LB_MED, UB_MED, NORMED_MED, PERF_MEAN, LB_MEAN, UB_MEAN, NORMED_MEAN),
        (ITER, METHOD, TEST_CASE),
    )
    baseline_mean_da = xru.only_dataarray(xru.ds_like(perf_da, ["ref"], (ITER, TEST_CASE)))
    # Using values here since just clearer to get raw items than xr object for func_name
    for func_name in perf_da.coords[TEST_CASE].values:
        rand_perf_med = baseline_ds[PERF_MED].sel({TEST_CASE: func_name}, drop=True).values
        rand_perf_mean = baseline_ds[PERF_MEAN].sel({TEST_CASE: func_name}, drop=True).values
        best_opt = baseline_ds[PERF_BEST].sel({TEST_CASE: func_name}, drop=True).values
        base_clip_val = baseline_ds[PERF_CLIP].sel({TEST_CASE: func_name}, drop=True).values

        assert np.all(np.diff(rand_perf_med) <= 0), "Baseline should be decreasing with iteration"
        assert np.all(np.diff(rand_perf_mean) <= 0), "Baseline should be decreasing with iteration"
        assert np.all(rand_perf_med > best_opt)
        assert np.all(rand_perf_mean > best_opt)
        assert np.all(rand_perf_mean <= base_clip_val)

        baseline_mean_da.loc[{TEST_CASE: func_name}] = linear_rescale(
            rand_perf_mean, best_opt, base_clip_val, 0.0, 1.0, enforce_bounds=False
        )
        for method_name in perf_da.coords[METHOD].values:
            # Take the minimum over all suggestion at given iter + sanity check perf_da
            curr_da = perf_da.sel({METHOD: method_name, TEST_CASE: func_name}, drop=True)
            assert curr_da.dims == (ITER, SUGGEST, TRIAL)

            if visible_perf_da is None:
                perf_array = get_perf_array(curr_da.values, curr_da.values)

                curr_da_ = perf_da.sel({METHOD: method_name, TEST_CASE: func_name}, drop=True).min(dim=SUGGEST)
                assert curr_da_.dims == (ITER, TRIAL)
                perf_array_ = np.minimum.accumulate(curr_da_.values, axis=0)
                assert np.allclose(perf_array, perf_array_)
            else:
                curr_visible_da = visible_perf_da.sel({METHOD: method_name, TEST_CASE: func_name}, drop=True)
                assert curr_visible_da.dims == (ITER, SUGGEST, TRIAL)
                perf_array = get_perf_array(curr_da.values, curr_visible_da.values)

            # Compute median perf and CI on it
            med_perf, LB, UB = qt.quantile_and_CI(perf_array, EVAL_Q, alpha=ALPHA)
            assert med_perf.shape == rand_perf_med.shape
            agg_result[PERF_MED].loc[{TEST_CASE: func_name, METHOD: method_name}] = med_perf
            agg_result[LB_MED].loc[{TEST_CASE: func_name, METHOD: method_name}] = LB
            agg_result[UB_MED].loc[{TEST_CASE: func_name, METHOD: method_name}] = UB

            # Now store normed version, which is better for aggregation
            normed = linear_rescale(med_perf, best_opt, rand_perf_med, 0.0, 1.0, enforce_bounds=False)
            agg_result[NORMED_MED].loc[{TEST_CASE: func_name, METHOD: method_name}] = normed

            # Store normed mean version
            normed = linear_rescale(perf_array, best_opt, base_clip_val, 0.0, 1.0, enforce_bounds=False)
            # Also, clip the score from below at -1 to limit max influence of single run on final average
            normed = np.clip(normed, -1.0, 1.0)
            normed = np.mean(normed, axis=1)
            agg_result[NORMED_MEAN].loc[{TEST_CASE: func_name, METHOD: method_name}] = normed

            # Compute mean perf and CI on it
            perf_array = np.minimum(base_clip_val, perf_array)
            mean_perf = np.mean(perf_array, axis=1)
            assert mean_perf.shape == rand_perf_mean.shape
            EB = t_EB(perf_array, alpha=ALPHA, axis=1)
            assert EB.shape == rand_perf_mean.shape
            agg_result[PERF_MEAN].loc[{TEST_CASE: func_name, METHOD: method_name}] = mean_perf
            agg_result[LB_MEAN].loc[{TEST_CASE: func_name, METHOD: method_name}] = mean_perf - EB
            agg_result[UB_MEAN].loc[{TEST_CASE: func_name, METHOD: method_name}] = mean_perf + EB
    assert not any(np.any(np.isnan(agg_result[kk].values)) for kk in agg_result)

    # Compute summary score over all test cases, summarize performance of each method
    summary = xru.ds_like(
        perf_da,
        (PERF_MED, LB_MED, UB_MED, PERF_MEAN, LB_MEAN, UB_MEAN, NORMED_MEAN, LB_NORMED_MEAN, UB_NORMED_MEAN),
        (ITER, METHOD),
    )
    summary[PERF_MED], summary[LB_MED], summary[UB_MED] = xr.apply_ufunc(
        qt.quantile_and_CI,
        agg_result[NORMED_MED],
        input_core_dims=[[TEST_CASE]],
        kwargs={"q": EVAL_Q, "alpha": ALPHA},
        output_core_dims=[[], [], []],
    )

    summary[PERF_MEAN] = agg_result[NORMED_MEAN].mean(dim=TEST_CASE)
    EB = xr.apply_ufunc(t_EB, agg_result[NORMED_MEAN], input_core_dims=[[TEST_CASE]])
    summary[LB_MEAN] = summary[PERF_MEAN] - EB
    summary[UB_MEAN] = summary[PERF_MEAN] + EB

    normalizer = baseline_mean_da.mean(dim=TEST_CASE)
    summary[NORMED_MEAN] = summary[PERF_MEAN] / normalizer
    summary[LB_NORMED_MEAN] = summary[LB_MEAN] / normalizer
    summary[UB_NORMED_MEAN] = summary[UB_MEAN] / normalizer

    assert all(tuple(summary[kk].dims) == (ITER, METHOD) for kk in summary)
    return agg_result, summary


def main():
    """See README for instructions on calling analysis.
    """
    description = "Analyze results from aggregated studies"
    args = parse_args(general_parser(description))

    # Metric used on leaderboard
    leaderboard_metric = cc.VISIBLE_TO_OPT

    logger.setLevel(logging.INFO)  # Note this is the module-wide logger
    if args[CmdArgs.verbose]:
        logger.addHandler(logging.StreamHandler())

    # Load in the eval data and sanity check
    perf_ds, meta = XRSerializer.load_derived(args[CmdArgs.db_root], db=args[CmdArgs.db], key=cc.EVAL_RESULTS)
    logger.info("Meta data from source file: %s" % str(meta["args"]))

    # Check if there is baselines file, other make one
    if cc.BASELINE not in XRSerializer.get_derived_keys(args[CmdArgs.db_root], db=args[CmdArgs.db]):
        warnings.warn("Baselines not found. Need to construct baseline.")
        do_baseline(args)

    # Load in baseline scores data and sanity check (including compatibility with eval data)
    baseline_ds, meta_ref = XRSerializer.load_derived(args[CmdArgs.db_root], db=args[CmdArgs.db], key=cc.BASELINE)
    logger.info("baseline data from source ref file: %s" % str(meta_ref["args"]))

    # Check test case signatures match between eval data and baseline data
    sig_errs, signatures = analyze_signature_pair(meta["signature"], meta_ref["signature"])
    logger.info("Signature errors:\n%s" % sig_errs.to_string())
    print(json.dumps({"exp-anal sig errors": sig_errs.T.to_dict()}))

    # Subset baseline to only the test cases run in the experiments
    test_cases_run = perf_ds.coords[TEST_CASE].values.tolist()
    assert set(test_cases_run) <= set(
        baseline_ds.coords[TEST_CASE].values.tolist()
    ), "Data set contains test cases not found in baseline."
    baseline_ds = baseline_ds.sel({TEST_CASE: test_cases_run})

    # Do the actual computation
    perf_visible = perf_ds[cc.VISIBLE_TO_OPT]
    agg_result = OrderedDict()
    summary = OrderedDict()
    for metric_for_scoring in sorted(perf_ds):
        perf_da = perf_ds[metric_for_scoring]
        baseline_ds_ = baseline_ds.sel({OBJECTIVE: metric_for_scoring}, drop=True)
        agg_result[(metric_for_scoring,)], summary[(metric_for_scoring,)] = compute_aggregates(
            perf_da, baseline_ds_, perf_visible
        )
    agg_result = xru.ds_concat(agg_result, dims=(cc.OBJECTIVE,))
    summary = xru.ds_concat(summary, dims=(cc.OBJECTIVE,))

    for metric_for_scoring in sorted(perf_ds):
        # Print summary by problem
        # Recall that:
        # ... summary[PERF_MEAN] = agg_result[NORMED_MEAN].mean(dim=TEST_CASE)
        # ... summary[NORMED_MEAN] = summary[PERF_MEAN] / normalizer
        # Where normalizer is constant across all problems, optimizers
        print("Scores by problem (JSON):\n")
        agg_df = agg_result[NORMED_MEAN].sel({cc.OBJECTIVE: metric_for_scoring}, drop=True)[{ITER: -1}].to_pandas().T
        print(json.dumps({metric_for_scoring: agg_df.to_dict()}))
        print("\n")

        final_score = summary[PERF_MED].sel({cc.OBJECTIVE: metric_for_scoring}, drop=True)[{ITER: -1}]
        logger.info("median score @ %d:\n%s" % (summary.sizes[ITER], xru.da_to_string(final_score)))
        final_score = summary[PERF_MEAN].sel({cc.OBJECTIVE: metric_for_scoring}, drop=True)[{ITER: -1}]
        logger.info("mean score @ %d:\n%s" % (summary.sizes[ITER], xru.da_to_string(final_score)))

        print("Final scores (JSON):\n")
        print(json.dumps({metric_for_scoring: final_score.to_series().to_dict()}))
        print("\n")

        final_score = summary[NORMED_MEAN].sel({cc.OBJECTIVE: metric_for_scoring}, drop=True)[{ITER: -1}]
        logger.info("normed mean score @ %d:\n%s" % (summary.sizes[ITER], xru.da_to_string(final_score)))

    # Now saving results
    meta = {"args": serializable_dict(args), "signature": signatures}
    XRSerializer.save_derived(agg_result, meta, args[CmdArgs.db_root], db=args[CmdArgs.db], key=cc.PERF_RESULTS)

    XRSerializer.save_derived(summary, meta, args[CmdArgs.db_root], db=args[CmdArgs.db], key=cc.MEAN_SCORE)

    final_msg = xru.da_to_string(
        100 * (1.0 - summary[PERF_MEAN].sel({cc.OBJECTIVE: leaderboard_metric}, drop=True)[{ITER: -1}])
    )
    logger.info("-" * 20)
    logger.info("Final score `100 x (1-loss)` for leaderboard:\n%s" % final_msg)


if __name__ == "__main__":
    main()  # pragma: main



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from copy import copy

import numpy as np
from poap.strategy import EvalRecord
from pySOT.experimental_design import SymmetricLatinHypercube
from pySOT.optimization_problems import OptimizationProblem
from pySOT.strategy import SRBFStrategy
from pySOT.surrogate import CubicKernel, LinearTail, RBFInterpolant, SurrogateUnitBox

from bayesmark.abstract_optimizer import AbstractOptimizer
from bayesmark.space import JointSpace


class PySOTOptimizer(AbstractOptimizer):
    primary_import = "pysot"

    def __init__(self, api_config, **kwargs):
        """Build wrapper class to use an optimizer in benchmark.

        Parameters
        ----------
        api_config : dict-like of dict-like
            Configuration of the optimization variables. See API description.
        """
        AbstractOptimizer.__init__(self, api_config)

        self.space_x = JointSpace(api_config)
        self.bounds = self.space_x.get_bounds()
        self.create_opt_prob()  # Sets up the optimization problem (needs self.bounds)
        self.max_evals = np.iinfo(np.int32).max  # NOTE: Largest possible int
        self.batch_size = None
        self.history = []
        self.proposals = []

    def create_opt_prob(self):
        """Create an optimization problem object."""
        opt = OptimizationProblem()
        opt.lb = self.bounds[:, 0]  # In warped space
        opt.ub = self.bounds[:, 1]  # In warped space
        opt.dim = len(self.bounds)
        opt.cont_var = np.arange(len(self.bounds))
        opt.int_var = []
        assert len(opt.cont_var) + len(opt.int_var) == opt.dim
        opt.objfun = None
        self.opt = opt

    def start(self, max_evals):
        """Starts a new pySOT run."""
        self.history = []
        self.proposals = []

        # Symmetric Latin hypercube design
        des_pts = max([self.batch_size, 2 * (self.opt.dim + 1)])
        slhd = SymmetricLatinHypercube(dim=self.opt.dim, num_pts=des_pts)

        # Warped RBF interpolant
        rbf = RBFInterpolant(dim=self.opt.dim, kernel=CubicKernel(), tail=LinearTail(self.opt.dim), eta=1e-4)
        rbf = SurrogateUnitBox(rbf, lb=self.opt.lb, ub=self.opt.ub)

        # Optimization strategy
        self.strategy = SRBFStrategy(
            max_evals=self.max_evals,
            opt_prob=self.opt,
            exp_design=slhd,
            surrogate=rbf,
            asynchronous=True,
            batch_size=1,
            use_restarts=True,
        )

    def suggest(self, n_suggestions=1):
        """Get a suggestion from the optimizer.

        Parameters
        ----------
        n_suggestions : int
            Desired number of parallel suggestions in the output

        Returns
        -------
        next_guess : list of dict
            List of `n_suggestions` suggestions to evaluate the objective
            function. Each suggestion is a dictionary where each key
            corresponds to a parameter being optimized.
        """

        if self.batch_size is None:  # First call to suggest
            self.batch_size = n_suggestions
            self.start(self.max_evals)

        # Set the tolerances pretending like we are running batch
        d, p = float(self.opt.dim), float(n_suggestions)
        self.strategy.failtol = p * int(max(np.ceil(d / p), np.ceil(4 / p)))

        # Now we can make suggestions
        x_w = []
        self.proposals = []
        for _ in range(n_suggestions):
            proposal = self.strategy.propose_action()
            record = EvalRecord(proposal.args, status="pending")
            proposal.record = record
            proposal.accept()  # This triggers all the callbacks

            # It is possible that pySOT proposes a previously evaluated point
            # when all variables are integers, so we just abort in this case
            # since we have likely converged anyway. See PySOT issue #30.
            x = list(proposal.record.params)  # From tuple to list
            x_unwarped, = self.space_x.unwarp(x)
            if x_unwarped in self.history:
                warnings.warn("pySOT proposed the same point twice")
                self.start(self.max_evals)
                return self.suggest(n_suggestions=n_suggestions)

            # NOTE: Append unwarped to avoid rounding issues
            self.history.append(copy(x_unwarped))
            self.proposals.append(proposal)
            x_w.append(copy(x_unwarped))

        return x_w

    def _observe(self, x, y):
        # Find the matching proposal and execute its callbacks
        idx = [x == xx for xx in self.history]
        i = np.argwhere(idx)[0].item()  # Pick the first index if there are ties
        proposal = self.proposals[i]
        proposal.record.complete(y)
        self.proposals.pop(i)
        self.history.pop(i)

    def observe(self, X, y):
        """Send an observation of a suggestion back to the optimizer.

        Parameters
        ----------
        X : list of dict-like
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,)
            Corresponding values where objective has been evaluated
        """
        assert len(X) == len(y)

        for x_, y_ in zip(X, y):
            # Just ignore, any inf observations we got, unclear if right thing
            if np.isfinite(y_):
                self._observe(x_, y_)


opt_wrapper = PySOTOptimizer



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General constants that are used in multiple files in the code base.
"""

# Special constant for random search since it gets used as our reference point in the baselines
RANDOM_SEARCH = "RandomSearch"
OPTIMIZERS_FILE = "config.json"
ARG_DELIM = "_"  # Delimeter used when concat cmd argument for any reason
PY_INTERPRETER = "python"  # What command to call for sub process, we could specify version number here also.

# Variables to save in SAL
EVAL = "eval"
TIME = "time"
SUGGEST_LOG = "suggest_log"
EXP_VARS = (EVAL, TIME, SUGGEST_LOG)

# Derived variables to save in SAL
TIME_RESULTS = "time"
EVAL_RESULTS = "eval"
BASELINE = "baseline"
PERF_RESULTS = "perf"
MEAN_SCORE = "summary"

# Coordinate dim names needed in saved xr Datasets
ITER = "iter"
TEST_CASE = "function"
METHOD = "optimizer"
TRIAL = "study_id"
SUGGEST = "suggestion"
OBJECTIVE = "objective"

# Dataset variables for eval results
VISIBLE_TO_OPT = "_visible_to_opt"

# Dataset variables for time results
SUGGEST_PHASE = "suggest"
OBS_PHASE = "observe"
EVAL_PHASE = "eval"
EVAL_PHASE_SUM = "eval_sum"
EVAL_PHASE_MAX = "eval_max"

# Dataset variables for aggregate results
PERF_MED = "median"
LB_MED = "median LB"
UB_MED = "median UB"
NORMED_MED = "median normed"
PERF_MEAN = "mean"
LB_MEAN = "mean LB"
UB_MEAN = "mean UB"
NORMED_MEAN = "mean normed"
LB_NORMED_MEAN = "mean normed LB"
UB_NORMED_MEAN = "mean normed UB"
PERF_BEST = "best"
PERF_CLIP = "clip"

# Choices used for test problems, there is some redundant specification with sklearn funcs file here
MODEL_NAMES = ("DT", "MLP-adam", "MLP-sgd", "RF", "SVM", "ada", "kNN", "lasso", "linear")
DATA_LOADER_NAMES = ("breast", "digits", "iris", "wine", "boston", "diabetes")

SCORERS_CLF = ("nll", "acc")
SCORERS_REG = ("mae", "mse")
METRICS = tuple(sorted(SCORERS_CLF + SCORERS_REG))



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module to deal with all matters relating to loading example data sets, which we tune ML models to.
"""
from enum import IntEnum, auto

import numpy as np
import pandas as pd  # only needed for csv reader, maybe try something else
from sklearn import datasets

from bayesmark.constants import DATA_LOADER_NAMES, SCORERS_CLF, SCORERS_REG
from bayesmark.path_util import join_safe_r
from bayesmark.stats import robust_standardize


class ProblemType(IntEnum):
    """The different problem types we consider. Currently, just regression (`reg`) and classification (`clf`).
    """

    clf = auto()
    reg = auto()


DATA_LOADERS = {
    "digits": (datasets.load_digits, ProblemType.clf),
    "iris": (datasets.load_iris, ProblemType.clf),
    "wine": (datasets.load_wine, ProblemType.clf),
    "breast": (datasets.load_breast_cancer, ProblemType.clf),
    "boston": (datasets.load_boston, ProblemType.reg),
    "diabetes": (datasets.load_diabetes, ProblemType.reg),
}

assert sorted(DATA_LOADERS.keys()) == sorted(DATA_LOADER_NAMES)

# Arguably, this could go in constants, but doesn't cause extra imports being here.
METRICS_LOOKUP = {ProblemType.clf: SCORERS_CLF, ProblemType.reg: SCORERS_REG}


def get_problem_type(dataset_name):
    """Determine if this dataset is a regression of classification problem.

    Parameters
    ----------
    dataset : str
        Which data set to use, must be key in `DATA_LOADERS` dict, or name of custom csv file.

    Returns
    -------
    problem_type : ProblemType
        `Enum` to indicate if regression of classification data set.
    """
    if dataset_name in DATA_LOADERS:
        _, problem_type = DATA_LOADERS[dataset_name]
        return problem_type

    # Maybe we can come up with a better system, but for now let's use a convention based on the naming of the csv file.
    if dataset_name.startswith("reg-"):
        return ProblemType.reg
    if dataset_name.startswith("clf-"):
        return ProblemType.clf
    assert False, "Can't determine problem type from dataset name."


def _csv_loader(dataset_name, return_X_y, data_root, clip_x=100):  # pragma: io
    """Load custom csv files for use in the benchmark.

    This function assumes ``dataset_name + ".csv"`` is a csv file found in the `data_root` path.  It also assumes the
    last column of the csv file is the target and the other columns are features.

    The target column should be `int` for classification and `float` for regression. Column names ending in ``"_cat"``
    are assumed to be categorical and will be one-hot encoded.

    The features (and target for regression) are robust standardized. The features are also clipped to be in
    ``[-clip_x, clip_x]`` *after* standardization.
    """
    assert return_X_y, "Only returning (X,y) tuple supported right now."
    assert clip_x >= 0

    # Quantile range for robust standardization. The 86% range is the most efficient for Gaussians. See:
    # https://github.com/scikit-learn/scikit-learn/issues/10139#issuecomment-344705040
    q_level = 0.86

    path = join_safe_r(data_root, dataset_name + ".csv")

    # For now, use convention that can get problem type based on data set name
    problem_type = get_problem_type(dataset_name)

    # Assuming no missing data in source csv files at the moment, these will
    # result in error.
    df = pd.read_csv(
        path, header=0, index_col=False, engine="c", na_filter=False, true_values=["true"], false_values=["false"]
    )

    label = df.columns[-1]  # Assume last col is target

    target = df.pop(label).values
    if problem_type == ProblemType.clf:
        assert target.dtype in (np.bool_, np.int_)
        target = target.astype(np.int_)  # convert to int for skl
    if problem_type == ProblemType.reg:
        assert target.dtype == np.float_
        # 86% range is the most efficient (at least for Gaussians)
        target = robust_standardize(target, q_level=q_level)

    # Fill in an categorical variables (object dtype of cols names ..._cat)
    cat_cols = sorted(cc for cc in df.columns if cc.endswith("_cat") or df[cc].dtype.kind == "O")
    df = pd.get_dummies(df, columns=cat_cols, drop_first=True, dtype=np.float_)
    # Could also sort all columns to be sure it will be reprod

    # Everything should now be in float
    assert (df.dtypes == np.float_).all()

    data = df.values
    data = robust_standardize(data, q_level=q_level)
    # Debatable if we should include this, but there are a lot of outliers
    data = np.clip(data, -clip_x, clip_x)

    # We should probably do some logging or something to wrap up
    return data, target, problem_type


def load_data(dataset_name, data_root=None):  # pragma: io
    """Load a data set and return it in, pre-processed into numpy arrays.

    Parameters
    ----------
    dataset : str
        Which data set to use, must be key in `DATA_LOADERS` dict, or name of custom csv file.
    data_root : str
        Root directory to look for all custom csv files. May be ``None`` for sklearn data sets.

    Returns
    -------
    data : :class:`numpy:numpy.ndarray` of shape (n, d)
        The feature matrix of the data set. It will be `float` array.
    target : :class:`numpy:numpy.ndarray` of shape (n,)
        The target vector for the problem, which is `int` for classification and `float` for regression.
    problem_type : :class:`bayesmark.data.ProblemType`
        `Enum` to indicate if regression of classification data set.
    """
    if dataset_name in DATA_LOADERS:
        loader_f, problem_type = DATA_LOADERS[dataset_name]
        data, target = loader_f(return_X_y=True)
    else:  # try to load as custom csv
        assert data_root is not None, "data root cannot be None when custom csv requested."
        data, target, problem_type = _csv_loader(dataset_name, return_X_y=True, data_root=data_root)
    return data, target, problem_type



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import find_packages, setup

CMD_NAME = "bayesmark"

# Strings to remove from README to make it PyPI friendly. See:
# https://packaging.python.org/guides/making-a-pypi-friendly-readme/#validating-restructuredtext-markup
REMOVE_FROM_RST = (":func:", ":ref:")


def read_requirements(name):
    with open("requirements/" + name + ".in") as f:
        requirements = f.read().strip()
    requirements = requirements.replace("==", ">=").splitlines()  # Loosen strict pins
    return [pp for pp in requirements if pp[0].isalnum()]


# Derive install requires from base.in first order requirements
requirements = read_requirements("base")
opt_requirements = read_requirements("optimizers")
ipynb_requirements = read_requirements("ipynb")

with open("README.rst") as f:
    long_description = f.read()
# Probably more efficient way to do this with regex but good enough
for remove_word in REMOVE_FROM_RST:
    long_description = long_description.replace(remove_word, "")

setup(
    name="bayesmark",
    version="0.0.5",
    packages=find_packages(),
    url="https://github.com/uber/bayesmark/",
    author="Ryan Turner",
    author_email=("rdturnermtl@github.com"),
    license="Apache v2",
    description="Bayesian optimization benchmark system",
    install_requires=requirements,
    extras_require={"optimizers": opt_requirements, "notebooks": ipynb_requirements},
    long_description=long_description,
    long_description_content_type="text/x-rst",
    platforms=["any"],
    entry_points={
        "console_scripts": [
            CMD_NAME + "-init = bayesmark.experiment_db_init:main",
            CMD_NAME + "-launch = bayesmark.experiment_launcher:main",
            CMD_NAME + "-agg = bayesmark.experiment_aggregate:main",
            CMD_NAME + "-baseline = bayesmark.experiment_baseline:main",
            CMD_NAME + "-anal = bayesmark.experiment_analysis:main",
            CMD_NAME + "-exp = bayesmark.experiment:main",
        ]
    },
)



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from scipy.interpolate import interp1d
from skopt import Optimizer as SkOpt
from skopt.space import Categorical, Integer, Real

from bayesmark.abstract_optimizer import AbstractOptimizer


class ScikitOptimizer(AbstractOptimizer):
    primary_import = "scikit-optimize"

    def __init__(self, api_config, base_estimator="GP", acq_func="gp_hedge", n_initial_points=5, **kwargs):
        """Build wrapper class to use an optimizer in benchmark.

        Parameters
        ----------
        api_config : dict-like of dict-like
            Configuration of the optimization variables. See API description.
        base_estimator : {'GP', 'RF', 'ET', 'GBRT'}
            How to estimate the objective function.
        acq_func : {'LCB', 'EI', 'PI', 'gp_hedge', 'EIps', 'PIps'}
            Acquisition objective to decide next suggestion.
        n_initial_points : int
            Number of points to sample randomly before actual Bayes opt.
        """
        AbstractOptimizer.__init__(self, api_config)

        dimensions, self.round_to_values = ScikitOptimizer.get_sk_dimensions(api_config)

        # Older versions of skopt don't copy over the dimensions names during
        # normalization and hence the names are missing in
        # self.skopt.space.dimensions. Therefore, we save our own copy of
        # dimensions list to be safe. If we can commit to using the newer
        # versions of skopt we can delete self.dimensions.
        self.dimensions_list = tuple(dd.name for dd in dimensions)

        # Undecided where we want to pass the kwargs, so for now just make sure
        # they are blank
        assert len(kwargs) == 0

        self.skopt = SkOpt(
            dimensions,
            n_initial_points=n_initial_points,
            base_estimator=base_estimator,
            acq_func=acq_func,
            acq_optimizer="auto",
            acq_func_kwargs={},
            acq_optimizer_kwargs={},
        )

    @staticmethod
    def get_sk_dimensions(api_config, transform="normalize"):
        """Help routine to setup skopt search space in constructor.

        Take api_config as argument so this can be static.
        """
        # The ordering of iteration prob makes no difference, but just to be
        # safe and consistnent with space.py, I will make sorted.
        param_list = sorted(api_config.keys())

        sk_dims = []
        round_to_values = {}
        for param_name in param_list:
            param_config = api_config[param_name]

            param_type = param_config["type"]

            param_space = param_config.get("space", None)
            param_range = param_config.get("range", None)
            param_values = param_config.get("values", None)

            # Some setup for case that whitelist of values is provided:
            values_only_type = param_type in ("cat", "ordinal")
            if (param_values is not None) and (not values_only_type):
                assert param_range is None
                param_values = np.unique(param_values)
                param_range = (param_values[0], param_values[-1])
                round_to_values[param_name] = interp1d(
                    param_values, param_values, kind="nearest", fill_value="extrapolate"
                )

            if param_type == "int":
                # Integer space in sklearn does not support any warping => Need
                # to leave the warping as linear in skopt.
                sk_dims.append(Integer(param_range[0], param_range[-1], transform=transform, name=param_name))
            elif param_type == "bool":
                assert param_range is None
                assert param_values is None
                sk_dims.append(Integer(0, 1, transform=transform, name=param_name))
            elif param_type in ("cat", "ordinal"):
                assert param_range is None
                # Leave x-form to one-hot as per skopt default
                sk_dims.append(Categorical(param_values, name=param_name))
            elif param_type == "real":
                # Skopt doesn't support all our warpings, so need to pick
                # closest substitute it does support.
                prior = "log-uniform" if param_space in ("log", "logit") else "uniform"
                sk_dims.append(Real(param_range[0], param_range[-1], prior=prior, transform=transform, name=param_name))
            else:
                assert False, "type %s not handled in API" % param_type
        return sk_dims, round_to_values

    def suggest(self, n_suggestions=1):
        """Get a suggestion from the optimizer.

        Parameters
        ----------
        n_suggestions : int
            Desired number of parallel suggestions in the output

        Returns
        -------
        next_guess : list of dict
            List of `n_suggestions` suggestions to evaluate the objective
            function. Each suggestion is a dictionary where each key
            corresponds to a parameter being optimized.
        """
        # First get list of lists from skopt.ask()
        next_guess = self.skopt.ask(n_points=n_suggestions)
        # Then convert to list of dicts
        next_guess = [dict(zip(self.dimensions_list, x)) for x in next_guess]

        # Now do the rounding, custom rounding is not supported in skopt. Note
        # that there is not nec a round function for each dimension here.
        for param_name, round_f in self.round_to_values.items():
            for xx in next_guess:
                xx[param_name] = round_f(xx[param_name])
        return next_guess

    def observe(self, X, y):
        """Send an observation of a suggestion back to the optimizer.

        Parameters
        ----------
        X : list of dict-like
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,)
            Corresponding values where objective has been evaluated
        """
        # Supposedly skopt can handle blocks, but not sure about interface for
        # that. Just do loop to be safe for now.
        for xx, yy in zip(X, y):
            # skopt needs lists instead of dicts
            xx = [xx[dim_name] for dim_name in self.dimensions_list]
            # Just ignore, any inf observations we got, unclear if right thing
            if np.isfinite(yy):
                self.skopt.tell(xx, yy)


opt_wrapper = ScikitOptimizer



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A serialization abstraction layer (SAL) to save and load experimental results. All IO of experimental results should
go through this module. This makes changing the backend (between different databases) transparent to the benchmark code.
"""
import json
import os
import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from tempfile import mkdtemp

import xarray as xr
from pathvalidate.argparse import validate_filename, validate_filepath

from bayesmark.path_util import join_safe_r, join_safe_w
from bayesmark.util import chomp, str_join_safe

NEWLINE = "\n"  # Just to be explicit, in case this ever gets run on Windows
PREFIX_FMT = "bo_%Y%m%d_%H%M%S_"  # The format we use for generating a new database name if none is specified

_XR_EXT = ".json"  # Extension we use for dumping xr.Dataset variables
_LOG_EXT = ".log"  # Extension to reccomend for logging files
_DERIVED_DIR = "derived"  # The folder for dervied variables (datasets)
_LOGGING_DIR = "log"  # The folder to reccomend for logging
_SETUP_STR = """
User must ensure
%s
exists, and setup folder using
mkdir %s
User must ensure equal reps of each optimizer for unbiased results."""


class Serializer(ABC):
    """Abstract base class for the serialization abstraction layer.
    """

    @staticmethod
    @abstractmethod
    def init_db(db_root, keys, db=None, exist_ok=True):
        """Initialize a "database" for storing data at the specified location.

        Parameters
        ----------
        db_root : str
            Absolute path to the database.
        keys : list(str)
            The variable names (or keys) we will store in the database for non-derived data.
        db : str
            The name of the database. If ``None``, a non-conflicting name will be generated.
        exist_ok : bool
            If true, do not raise an error if this database already exists.

        Returns
        -------
        db : str
            The name of the database.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_keys(db_root, db):
        """List the non-derived keys available in the database.

        Parameters
        ----------
        db_root : str
            Absolute path to the database.
        db : str
            The name of the database.

        Returns
        -------
        keys : list(str)
            The variable names (or keys) in the database for non-derived data.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_derived_keys(db_root, db):
        """List the derived keys currently available in the database.

        Parameters
        ----------
        db_root : str
            Absolute path to the database.
        db : str
            The name of the database.

        Returns
        -------
        keys : list(str)
            The variable names (or keys) in the database for derived data.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_uuids(db_root, db, key):
        """List the UUIDs for the versions of a variable (non-derived key) available in the database.

        Parameters
        ----------
        db_root : str
            Absolute path to the database.
        db : str
            The name of the database.
        keys : str
            The variable name in the database for non-derived data.

        Returns
        -------
        uuids : list(uuid.UUID)
            The UUIDs for the versions of this key.
        """
        pass

    @staticmethod
    @abstractmethod
    def save(data, meta, db_root, db, key, uuid_):
        """Abstract method for saving experimental data, details require the type of `data`.
        """
        pass

    @staticmethod
    @abstractmethod
    def load(db_root, db, key, uuid_):
        """Abstract method for loading experimental data, details require the type of `data`.
        """
        pass

    @staticmethod
    @abstractmethod
    def save_derived(data, meta, db_root, db, key):
        """Abstract method for saving derived data, details require the type of `data`.
        """
        pass

    @staticmethod
    @abstractmethod
    def load_derived(db_root, db, key):
        """Abstract method for loading derived data, details require the type of `data`.
        """
        pass


class XRSerializer(Serializer):
    """Serialization layer when saving and loading `xarray` datasets (currently) as `json`.
    """

    def init_db(db_root, keys, db=None, exist_ok=True):  # pragma: io
        XRSerializer._validate(db_root, keys, db)

        if db is None:
            folder_prefix = datetime.utcnow().strftime(PREFIX_FMT)
            exp_subdir = mkdtemp(prefix=folder_prefix, dir=db_root)
            db = os.path.basename(exp_subdir)
            assert db.startswith(folder_prefix)
            assert os.path.join(db_root, db) == exp_subdir
        else:
            exp_subdir = os.path.join(db_root, db)
            os.makedirs(exp_subdir, exist_ok=exist_ok)

        subdirs = [_DERIVED_DIR, _LOGGING_DIR] + list(keys)
        for subd in subdirs:
            os.makedirs(os.path.join(exp_subdir, subd), exist_ok=exist_ok)

        return db

    def init_db_manual(db_root, keys, db):
        """Instruction for how one would manually initialize the "database" on another system.

        Parameters
        ----------
        db_root : str
            Absolute path to the database.
        keys : list(str)
            The variable names (or keys) we will store in the database for non-derived data.
        db : str
            The name of the database.

        Returns
        -------
        manual_setup_info : str
            The setup instructions.
        """
        XRSerializer._validate(db_root, keys, db)
        assert db is not None, "Must specify db name to setup manually."

        exp_subdir = os.path.join(db_root, db)
        subdirs = [_DERIVED_DIR, _LOGGING_DIR] + list(keys)
        manual_setup_info = _SETUP_STR % (exp_subdir, str_join_safe(" ", subdirs))
        return manual_setup_info

    def get_keys(db_root, db):  # pragma: io
        XRSerializer._validate(db_root, keys=(), db=db)

        keys = sorted(os.listdir(os.path.join(db_root, db)))
        keys.remove(_DERIVED_DIR)
        keys.remove(_LOGGING_DIR)
        return keys

    def get_derived_keys(db_root, db):  # pragma: io
        XRSerializer._validate(db_root, keys=(), db=db)

        fnames = sorted(os.listdir(os.path.join(db_root, db, _DERIVED_DIR)))
        keys = [XRSerializer._fname_to_key(ff) for ff in fnames]
        return keys

    def get_uuids(db_root, db, key):  # pragma: io
        XRSerializer._validate(db_root, keys=[key], db=db)

        fnames = sorted(os.listdir(os.path.join(db_root, db, key)))
        uuids = [XRSerializer._fname_to_uuid(ff) for ff in fnames]
        return uuids

    def save(data, meta, db_root, db, key, uuid_):  # pragma: io
        """Save a dataset under a key name in the database.

        Parameters
        ----------
        data : :class:`xarray:xarray.Dataset`
            An :class:`xarray:xarray.Dataset` variable we would like to store as non-derived data from an experiment.
        meta : json-serializable
            Associated meta-data with the experiment. This can be anything json serializable.
        db_root : str
            Absolute path to the database.
        db : str
            The name of the database.
        key : str
            The variable name in the database for the data.
        uuid_ : uuid.UUID
            The UUID to represent the version of this variable we are storing.
        """
        XRSerializer._validate(db_root, keys=[key], db=db)

        fname = XRSerializer._uuid_to_fname(uuid_)
        path = (db_root, db, key, fname)
        with open(join_safe_w(*path), "w") as f:
            _dump_xr(f, ds=data, meta=meta)

    def load(db_root, db, key, uuid_):  # pragma: io
        """Load a dataset under a key name in the database. This is the inverse of :func:`.save`.

        Parameters
        ----------
        db_root : str
            Absolute path to the database.
        db : str
            The name of the database.
        key : str
            The variable name in the database for the data.
        uuid_ : uuid.UUID
            The UUID to represent the version of this variable we want to load.

        Returns
        -------
        data : :class:`xarray:xarray.Dataset`
            An :class:`xarray:xarray.Dataset` variable for the non-derived data from an experiment.
        meta : json-serializable
            Associated meta-data with the experiment. This can be anything json serializable.
        """
        XRSerializer._validate(db_root, keys=[key], db=db)

        fname = XRSerializer._uuid_to_fname(uuid_)
        path = (db_root, db, key, fname)
        with open(join_safe_r(*path), "r") as f:
            ds, meta = _load_xr(f)
        return ds, meta

    def save_derived(data, meta, db_root, db, key):  # pragma: io
        """Save a dataset under a key name in the database as derived data.

        Parameters
        ----------
        data : :class:`xarray:xarray.Dataset`
            An :class:`xarray:xarray.Dataset` variable we would like to store as derived data from experiments.
        meta : json-serializable
            Associated meta-data with the experiments. This can be anything json serializable.
        db_root : str
            Absolute path to the database.
        db : str
            The name of the database.
        key : str
            The variable name in the database for the data.
        """
        XRSerializer._validate(db_root, keys=[key], db=db)

        fname = XRSerializer._key_to_fname(key)
        path = (db_root, db, _DERIVED_DIR, fname)
        with open(join_safe_w(*path), "w") as f:
            _dump_xr(f, ds=data, meta=meta)

    def load_derived(db_root, db, key):  # pragma: io
        """Load a dataset under a key name in the database as derived data. This is the inverse of :func:`.save_derived`.

        Parameters
        ----------
        db_root : str
            Absolute path to the database.
        db : str
            The name of the database.
        key : str
            The variable name in the database for the data.

        Returns
        -------
        data : :class:`xarray:xarray.Dataset`
            An :class:`xarray:xarray.Dataset` variable for the derived data from experiments.
        meta : json-serializable
            Associated meta-data with the experiments. This can be anything json serializable.
        """
        XRSerializer._validate(db_root, keys=[key], db=db)

        fname = XRSerializer._key_to_fname(key)
        path = (db_root, db, _DERIVED_DIR, fname)
        with open(join_safe_r(*path), "r") as f:
            data, meta = _load_xr(f)
        return data, meta

    def logging_path(db_root, db, uuid_):  # pragma: io
        """Get an absolute path for logging from an experiment given its UUID.

        Parameters
        ----------
        db_root : str
            Absolute path to the database.
        db : str
            The name of the database.
        uuid_ : uuid.UUID
            The UUID to represent this experiment.

        Returns
        -------
        logfile : str
            Absolute path suitable for logging in this experiment.
        """
        XRSerializer._validate(db_root, keys=(), db=db)
        assert isinstance(uuid_, uuid.UUID)

        fname = uuid_.hex + _LOG_EXT
        logfile = join_safe_w(db_root, db, _LOGGING_DIR, fname)
        return logfile

    def _fname_to_uuid(fname):
        uuid_ = uuid.UUID(chomp(fname, _XR_EXT))
        return uuid_

    def _uuid_to_fname(uuid_):
        assert isinstance(uuid_, uuid.UUID)  # This can be eliminated once we use type hints

        fname = uuid_.hex + _XR_EXT
        return fname

    def _key_to_fname(key):
        fname = key + _XR_EXT
        return fname

    def _fname_to_key(fname):
        key = chomp(fname, _XR_EXT)
        return key

    def _validate(db_root, keys=(), db=None):
        validate_filepath(db_root, platform="auto")
        assert os.path.isabs(db_root), "db_root must be absolute path"

        if db is not None:
            validate_filename(db, platform="universal")

        for kk in keys:
            validate_filename(kk, platform="universal")


def _dump_xr(f, ds, meta):  # pragma: io
    """Helper routine to `XRSerializer.save` and `XRSerializer.save_derived`.
    """
    assert isinstance(ds, xr.Dataset)  # Requiring Dataset and not DataArray for now

    meta_json = json.dumps(meta)  # meta can be anything that json can handle
    # JSON dumps seems pretty good about escaping, but check to be sure
    assert NEWLINE not in meta_json

    # Built in json dumper doesn't allow us to only line break on top-level, so we manually do this for now
    f.write('{"meta": %s,' % meta_json)
    f.write(NEWLINE)
    f.write('"data": ')

    json.dump(ds.to_dict(), f)
    f.write("}")
    f.write(NEWLINE)


def _load_xr(f):  # pragma: io
    """Helper routine to `XRSerializer.load` and`XRSerializer.load_derived`.
    """
    all_json = json.load(f)
    meta = all_json.pop("meta")
    ds = xr.Dataset.from_dict(all_json.pop("data"))
    return ds, meta



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General utilities that should arguably be included in Python.
"""
import shlex


def in_or_none(x, L):
    """Check if item is in list of list is None."""
    return (L is None) or (x in L)


def all_unique(L):
    """Check if all elements in a list are unique.

    Parameters
    ----------
    L : list
        List we would like to check for uniqueness.

    Returns
    -------
    uniq : bool
        True if all elements in `L` are unique.
    """
    uniq = len(L) == len(set(L))
    return uniq


def strict_sorted(L):
    """Return a strictly sorted version of `L`. Therefore, this raises an error if `L` contains duplicates.

    Parameters
    ----------
    L : list
        List we would like to sort.

    Returns
    -------
    S : list
        Strictly sorted version of `L`.
    """
    assert all_unique(L), "Cannot strict sort because list contains duplicates."
    S = sorted(L)
    return S


def range_str(stop):
    """Version of ``range(stop)`` that instead returns strings that are zero padded so the entire iteration is of the
    same length.

    Parameters
    ----------
    stop : int
        Stop value equivalent to ``range(stop)``.

    Yields
    ------
    x : str
        String representation of integer zero padded so all items from this generator have the same ``len(x)``.
    """
    str_len = len(str(stop - 1))  # moot if stop=0

    def map_(x):
        ss = str(x).zfill(str_len)
        return x, ss

    G = map(map_, range(stop))
    return G


def str_join_safe(delim, str_vec, append=False):
    """Version of `str.join` that is guaranteed to be invertible.

    Parameters
    ----------
    delim : str
        Delimiter to join the strings.
    str_vec : list(str)
        List of strings to join. A `ValueError` is raised if `delim` is present in any of these strings.
    append : bool
        If true, assume the first element is already joined and we are appending to it. So, `str_vec[0]` can contain
        `delim`.

    Returns
    -------
    joined_str : str
        Joined version of `str_vec`, which is always recoverable with ``joined_str.split(delim)``.

    Examples
    --------
    Append is required because,

    .. code-block:: pycon

        ss = str_join_safe('_', ('foo', 'bar'))
        str_join_safe('_', (ss, 'baz', 'qux'))

    would fail because we are appending ``'baz'`` and ``'qux'`` to the already joined string ``ss = 'foo_bar'``.

    In this case, we use

    .. code-block:: pycon

        ss = str_join_safe('_', ('foo', 'bar'))
        str_join_safe('_', (ss, 'baz', 'qux'), append=True)
    """
    chk_vec = str_vec[1:] if append else str_vec

    for ss in chk_vec:
        if delim in ss:
            raise ValueError("%s cannot contain delimeter %s" % (ss, delim))

    joined_str = delim.join(str_vec)
    return joined_str


def shell_join(argv, delim=" "):
    """Join strings together in a way that is an inverse of `shlex` shell parsing into `argv`.

    Basically, if the resulting string is passed as a command line argument then `sys.argv` will equal `argv`.

    Parameters
    ----------
    argv : list(str)
        List of arguments to collect into command line string. It will be escaped accordingly.
    delim : str
        Whitespace delimiter to join the strings.

    Returns
    -------
    cmd : str
        Properly escaped and joined command line string.
    """
    vv = [shlex.quote(vv) for vv in argv]
    cmd = delim.join(vv)
    assert shlex.split(cmd) == list(argv)
    return cmd


def chomp(str_val, ext="\n"):
    """Chomp a suffix off a string.

    Parameters
    ----------
    str_val : str
        String we want to chomp off a suffix, e.g., ``"foo.log"``, and we want to chomp the file extension.
    ext : str
        The suffix we want to chomp. An error is raised if `str_val` doesn't end in `ext`.

    Returns
    -------
    chomped : str
        Version of `str_val` with `ext` removed from the end.
    """
    n = len(ext)
    assert n > 0

    chomped, ext_ = str_val[:-n], str_val[-n:]
    assert ext == ext_, "%s must end with %s" % (repr(str_val), repr(ext))
    return chomped


def preimage_func(f, x):
    """Pre-image a funcation at a set of input points.

    Parameters
    ----------
    f : typing.Callable
        The function we would like to pre-image. The output type must be hashable.
    x : typing.Iterable
        Input points we would like to evaluate `f`. `x` must be of a type acceptable by `f`.

    Returns
    -------
    D : dict(object, list(object))
        This dictionary maps the output of `f` to the list of `x` values that produce it.
    """
    D = {}
    for xx in x:
        D.setdefault(f(xx), []).append(xx)
    return D
